from __future__ import annotations

import copy
import time
import typing as t
from enum import Enum, unique

from starwhale.utils import now_str
from starwhale.base.mixin import ASDictMixin
from starwhale.core.dataset.type import Link


@unique
class _TrackMode(Enum):
    OFFLINE = "offline"
    ONLINE = "online"


@unique
class _TrackType(Enum):
    METRICS = "metrics"
    PARAMETERS = "parameters"
    ARTIFACTS = "artifacts"


@unique
class _TrackSource(Enum):
    USER = "user"
    SYSTEM = "_system"
    FRAMEWORK = "_framework"


class TrackRecord:
    def __init__(
        self, typ: _TrackType, source: _TrackSource, start_time: float = 0
    ) -> None:
        self.typ = typ
        self.source = source
        self.clock_time = now_str()
        self.relative_time = time.monotonic() - start_time

    def __str__(self) -> str:
        return f"TrackRecord[{self.typ}] from {self.source}"


class MetricsRecord(TrackRecord):
    def __init__(
        self,
        data: t.Dict[str, float],
        step: int,
        source: _TrackSource = _TrackSource.USER,
        start_time: float = 0,
    ) -> None:
        super().__init__(_TrackType.METRICS, source, start_time)
        self.data = data
        self.step = step

    def __str__(self) -> str:
        return f"MetricsRecord[{self.step}] {self.data}"

    __repr__ = __str__


class ParamsRecord(TrackRecord):
    def __init__(
        self,
        data: t.Dict[str, t.Any],
        source: _TrackSource = _TrackSource.USER,
        start_time: float = 0,
    ) -> None:
        super().__init__(_TrackType.PARAMETERS, source, start_time)
        self.data = data

    def __str__(self) -> str:
        return f"ParamsRecord-{self.data}"

    __repr__ = __str__


class ArtifactsRecord(TrackRecord):
    def __init__(
        self,
        data: t.Dict[str, t.Any],
        source: _TrackSource = _TrackSource.USER,
        start_time: float = 0,
    ) -> None:
        super().__init__(_TrackType.ARTIFACTS, source, start_time)
        self.data = data

    def __str__(self) -> str:
        return f"ArtifactsRecord-{self.data}"

    __repr__ = __str__


class ArtifactsTabularRow(ASDictMixin):
    def __init__(
        self,
        name: str,
        index: int,
        created_at: str,
        data: Link,
        link_wrapper: bool = True,
    ) -> None:
        self.name = name
        self.index = index
        self.key = f"{name}-{index:04}"
        self.created_at = created_at
        self.data = data
        self._link_wrapper = link_wrapper

        self._do_validate()

    def _do_validate(self) -> None:
        if not isinstance(self.data, Link):
            raise TypeError(f"data:{self.data} is not Link type")

        if self.index < 0 or self.index >= 10000:
            raise ValueError(
                f"index({self.index}) must be non negative integer number and less than 10000"
            )

    def __str__(self) -> str:
        return f"ArtifactsTabularRow-{self.key}"

    def __repr__(self) -> str:
        return f"ArtifactsTabularRow-{self.key}-{self.created_at}-{self.data}"

    @classmethod
    def from_datastore(cls, **kw: t.Any) -> ArtifactsTabularRow:
        return cls(
            name=kw["name"],
            index=kw["index"],
            created_at=kw["created_at"],
            data=kw["data"],
            link_wrapper=kw.get("__link_wrapper", True),
        )

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        return {
            "__key": self.key,
            "__link_wrapper": self._link_wrapper,
            "name": self.name,
            "index": self.index,
            "data": self.data,
        }


class MetricsTabularRow(ASDictMixin):
    def __init__(
        self,
        step: int,
        clock_time: str,
        relative_time: float,
        metrics: t.Dict[str, float],
    ) -> None:
        self.step = step
        self.clock_time = clock_time
        self.relative_time = float(relative_time)
        self.metrics = metrics
        # TODO: add synced flag for syncer thread?
        self._do_validate()

    def _do_validate(self) -> None:
        if self.step < 0:
            raise TypeError(f"step({self.step}) should be non-negative integer number")

        for k, v in self.metrics.items():
            if not isinstance(k, str) or not isinstance(v, float):
                raise TypeError(
                    f"metric(k:{k}, v:{v}) should be str key and float value."
                )

    def __str__(self) -> str:
        return f"MetricsTabularRow[{self.step}] {self.metrics}"

    def __repr__(self) -> str:
        return f"MetricsTabularRow[{self.step}] {self.metrics} at {self.clock_time}, relative time({self.relative_time}s)"

    @classmethod
    def from_datastore(cls, **kw: t.Any) -> MetricsTabularRow:
        metrics: t.Dict[str, float] = {}

        for k, v in kw.items():
            if k.startswith("__"):
                continue

            if isinstance(k, str) and isinstance(v, (int, float)):
                metrics[k] = float(v)

        return cls(
            step=kw["__step"],
            clock_time=kw["__clock_time"],
            relative_time=kw["__relative_time"],
            metrics=metrics,
        )

    @classmethod
    def from_record(cls, record: MetricsRecord) -> MetricsTabularRow:
        return cls(
            step=record.step,
            clock_time=record.clock_time,
            relative_time=record.relative_time,
            metrics=record.data,
        )

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        ret: t.Dict[str, t.Any] = copy.deepcopy(self.metrics)
        ret.update(
            {
                "__step": self.step,
                "__clock_time": self.clock_time,
                "__relative_time": self.relative_time,
            }
        )
        return ret


HandleQueueElementType = t.Optional[
    t.Union[TrackRecord, MetricsRecord, ArtifactsRecord, ParamsRecord]
]



import typing as t
import os.path
from urllib.parse import urlparse

import httpx
import pkg_resources
from fastapi import FastAPI, APIRouter
from fastapi.responses import ORJSONResponse
from typing_extensions import Protocol
from starlette.requests import Request
from starlette.responses import Response, FileResponse, StreamingResponse
from starlette.background import BackgroundTask
from starlette.exceptions import HTTPException
from starlette.staticfiles import StaticFiles

from starwhale.web import user, panel, system, project, data_store
from starwhale.base.uricomponents.instance import Instance

STATIC_DIR_DEV = pkg_resources.resource_filename("starwhale", "web/ui")


class Component(Protocol):
    router: APIRouter
    prefix: str


class Server(FastAPI):
    def __init__(self) -> None:
        super().__init__(default_response_class=ORJSONResponse)

    def add_component(self, component: Component) -> None:
        self.include_router(component.router, prefix=f"/{component.prefix.lstrip('/')}")

    @classmethod
    def mount_static(cls, app: FastAPI) -> FastAPI:
        def index() -> FileResponse:
            return FileResponse(os.path.join(STATIC_DIR_DEV, "index.html"))

        app.add_route("/", index, methods=["GET"])
        app.mount("/", StaticFiles(directory=STATIC_DIR_DEV), name="assets")

        def not_found_exception_handler(
            request: Request, exc: HTTPException
        ) -> FileResponse:
            return FileResponse(os.path.join(STATIC_DIR_DEV, "index.html"))

        app.add_exception_handler(404, not_found_exception_handler)

        return app

    @classmethod
    def with_components(cls, components: t.List[Component]) -> FastAPI:
        api = Server()
        for component in components:
            api.add_component(component)

        app = FastAPI()
        app.mount("/api/v1", api)
        return cls.mount_static(app)

    @classmethod
    def default(cls) -> FastAPI:
        return cls.with_components([panel, project, data_store, user, system])

    @classmethod
    def proxy(cls, instance: Instance) -> FastAPI:
        app = FastAPI()
        client = httpx.AsyncClient(base_url=instance.url)
        host = urlparse(instance.url).netloc

        async def proxy(request: Request) -> Response:
            url = httpx.URL(path=request.url.path, params=request.query_params)
            headers = request.headers.mutablecopy()
            headers.update({"Authorization": instance.token, "Host": host})
            content = await request.body()
            req = client.build_request(
                request.method, url, content=content, headers=headers
            )
            resp = await client.send(req, stream=True)
            return StreamingResponse(
                resp.aiter_bytes(),
                status_code=resp.status_code,
                headers=resp.headers,
                background=BackgroundTask(resp.aclose),
            )

        # serve panel related api in local mode for customizing panel
        app.include_router(panel.router, prefix="/api/v1/panel")
        app.add_route(
            "/api/v1/{path:path}",
            proxy,
            methods=["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD"],
        )

        return cls.mount_static(app)


# for testing
if __name__ == "__main__":
    import uvicorn

    server = Server.default()
    uvicorn.run("server:server", host="127.0.0.1", port=8000, reload=True)



import os
import typing as t
from pathlib import Path

import click

from starwhale.consts import DefaultYAMLName, DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE
from starwhale.base.uri import URI
from starwhale.base.type import URIType
from starwhale.utils.cli import AliasedGroup
from starwhale.utils.load import import_object
from starwhale.utils.error import NotFoundError
from starwhale.core.dataset.type import DatasetAttr, DatasetConfig

from .view import get_term_view, DatasetTermView


@click.group(
    "dataset",
    cls=AliasedGroup,
    help="Dataset management, build/info/list/copy/tag...",
)
@click.pass_context
def dataset_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@dataset_cmd.command("build", help="[Only Standalone]Build swds with dataset.yaml")
@click.argument("workdir", type=click.Path(exists=True, file_okay=False))
@click.option(
    "-h",
    "--handler",
    help="Dataset build executor handler: [module path]:[class or function name]",
)
@click.option("-n", "--name", help="Dataset name")
@click.option(
    "-p",
    "--project",
    help="Project URI, the default is the current selected project. The dataset will store in the specified project",
)
@click.option("--desc", help="Dataset description")
@click.option(
    "-as",
    "--alignment-size",
    help="swds-bin format dataset: alignment size",
)
@click.option(
    "-vs",
    "--volume-size",
    help="swds-bin format dataset: volume size",
)
@click.option(
    "-f",
    "--dataset-yaml",
    default=DefaultYAMLName.DATASET,
    help="Dataset yaml filename, default use ${WORKDIR}/dataset.yaml file",
)
@click.option("-r", "--runtime", help="runtime uri")
@click.pass_obj
def _build(
    view: DatasetTermView,
    workdir: str,
    handler: str,
    name: str,
    project: str,
    desc: str,
    dataset_yaml: str,
    alignment_size: str,
    volume_size: str,
    runtime: str,
) -> None:
    # TODO: add dry-run
    # TODO: add compress args
    if not os.path.exists(workdir):
        raise NotFoundError(workdir)

    yaml_path = Path(workdir) / dataset_yaml
    config = DatasetConfig()
    if yaml_path.exists():
        config = DatasetConfig.create_by_yaml(yaml_path)

    config.name = name or config.name or Path(workdir).absolute().name
    handler = handler or config.handler
    config.handler = import_object(workdir, handler)
    config.runtime_uri = runtime or config.runtime_uri
    config.project_uri = project or config.project_uri
    # TODO: support README.md as the default desc
    config.desc = desc or config.desc

    config.attr = DatasetAttr(
        volume_size=volume_size or config.attr.volume_size,
        alignment_size=alignment_size or config.attr.alignment_size,
    )

    config.do_validate()
    view.build(workdir, config)


@dataset_cmd.command("diff", help="Dataset version diff")
@click.argument("base_uri", required=True)
@click.argument("compare_uri", required=True)
@click.option(
    "--show-details", is_flag=True, help="Show data different detail by the row"
)
@click.pass_obj
def _diff(
    view: t.Type[DatasetTermView], base_uri: str, compare_uri: str, show_details: bool
) -> None:
    view(base_uri).diff(URI(compare_uri, expected_type=URIType.DATASET), show_details)


@dataset_cmd.command("list", aliases=["ls"])
@click.option(
    "-p",
    "--project",
    default="",
    help="Project URI, the default is the current selected project.",
)
@click.option("-f", "--fullname", is_flag=True, help="Show fullname of dataset version")
@click.option("-sr", "--show-removed", is_flag=True, help="Show removed datasets")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for dataset list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for dataset list"
)
@click.option(
    "filters",
    "-fl",
    "--filter",
    multiple=True,
    help="Filter output based on conditions provided.",
)
@click.pass_obj
def _list(
    view: DatasetTermView,
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
    filters: list,
) -> None:
    """
    List Dataset of the specified project.

    The filtering flag (-fl or --filter) format is a key=value pair or a flag.
    If there is more than one filter, then pass multiple flags.\n
    (e.g. --filter name=mnist --filter latest)

    \b
    The currently supported filters are:
      name\tTEXT\tThe prefix of the dataset name
      owner\tTEXT\tThe name or id of the dataset owner
      latest\tFLAG\t[Cloud] Only show the latest version
            \t \t[Standalone] Only show the version with "latest" tag
    """
    view.list(project, fullname, show_removed, page, size, filters)


@dataset_cmd.command("info", help="Show dataset details")
@click.argument("dataset")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _info(view: t.Type[DatasetTermView], dataset: str, fullname: bool) -> None:
    view(dataset).info(fullname)


@dataset_cmd.command("remove", aliases=["rm"])
@click.argument("dataset")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force to remove dataset, the removed dataset cannot recover",
)
@click.pass_obj
def _remove(view: t.Type[DatasetTermView], dataset: str, force: bool) -> None:
    """
    Remove dataset

    You can run `swcli dataset recover` to recover the removed datasets.

    DATASET: argument use the `Dataset URI` format, so you can remove the whole dataset or a specified-version dataset.
    """
    click.confirm("continue to remove?", abort=True)
    view(dataset).remove(force)


@dataset_cmd.command("recover")
@click.argument("dataset")
@click.option("-f", "--force", is_flag=True, help="Force to recover dataset")
@click.pass_obj
def _recover(view: t.Type[DatasetTermView], dataset: str, force: bool) -> None:
    """
    Recover dataset

    DATASET: argument use the `Dataset URI` format, so you can recover the whole dataset or a specified-version dataset.
    """
    view(dataset).recover(force)


@dataset_cmd.command("history", help="Show dataset history")
@click.argument("dataset")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _history(
    view: t.Type[DatasetTermView], dataset: str, fullname: bool = False
) -> None:
    view(dataset).history(fullname)


@dataset_cmd.command("summary", help="Show dataset summary")
@click.argument("dataset")
@click.pass_obj
def _summary(view: t.Type[DatasetTermView], dataset: str) -> None:
    view(dataset).summary()


@dataset_cmd.command("copy", aliases=["cp"])
@click.argument("src")
@click.argument("dest")
@click.option("-dlp", "--dest-local-project", help="dest local project uri")
def _copy(src: str, dest: str, dest_local_project: str) -> None:
    """
    Copy Dataset between Standalone Instance and Cloud Instance

    SRC: dataset uri with version

    DEST: project uri or dataset uri with name.

    Example:

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud dataset to local project(myproject) with a new dataset name 'mnist-local'
            swcli dataset cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq local/project/myproject/mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud dataset to local default project(self) with the cloud instance dataset name 'mnist-cloud'
            swcli dataset cp cloud://pre-k8s/project/dataset/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq .

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud dataset to local project(myproject) with the cloud instance dataset name 'mnist-cloud'
            swcli dataset cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq . -dlp myproject

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud dataset to local default project(self) with a dataset name 'mnist-local'
            swcli dataset cp cloud://pre-k8s/project/dataset/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud dataset to local project(myproject) with a dataset name 'mnist-local'
            swcli dataset cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local -dlp myproject

        \b
        - copy standalone instance(local) default project(self)'s mnist-local dataset to cloud instance(pre-k8s) mnist project with a new dataset name 'mnist-cloud'
            swcli dataset cp mnist-local/version/latest cloud://pre-k8s/project/mnist/mnist-cloud

        \b
        - copy standalone instance(local) default project(self)'s mnist-local dataset to cloud instance(pre-k8s) mnist project with standalone instance dataset name 'mnist-local'
            swcli dataset cp mnist-local/version/latest cloud://pre-k8s/project/mnist

        \b
        - copy standalone instance(local) default project(self)'s mnist-local dataset to cloud instance(pre-k8s) mnist project without 'cloud://' prefix
            swcli dataset cp mnist-local/version/latest pre-k8s/project/mnist

        \b
        - copy standalone instance(local) project(myproject)'s mnist-local dataset to cloud instance(pre-k8s) mnist project with standalone instance dataset name 'mnist-local'
            swcli dataset cp local/project/myproject/dataset/mnist-local/version/latest cloud://pre-k8s/project/mnist
    """
    DatasetTermView.copy(src, dest, dest_local_project)


@dataset_cmd.command("tag", help="Dataset tag management, add or remove")
@click.argument("dataset")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
@click.pass_obj
def _tag(
    view: t.Type[DatasetTermView],
    dataset: str,
    tags: t.List[str],
    remove: bool,
    quiet: bool,
) -> None:
    view(dataset).tag(tags, remove, quiet)


@dataset_cmd.command("head", help="Print the first 5 rows of the dataset")
@click.argument("dataset")
@click.option("-n", "--rows", default=5, help="Print the first NUM rows of the dataset")
@click.option(
    "-d",
    "--show-raw-data",
    is_flag=True,
    help="Fetch raw data content",
)
@click.pass_obj
def _head(
    view: t.Type[DatasetTermView],
    dataset: str,
    rows: int,
    show_raw_data: bool,
) -> None:
    view(dataset).head(rows, show_raw_data)



from __future__ import annotations

import queue
import typing as t
import threading

from starwhale.base.uri import URI


class SyncerThread(threading.Thread):
    def __init__(self, sync_queue: queue.Queue[t.Any], project_uri: URI) -> None:
        super().__init__(name=f"SyncerThread-{project_uri}")

        self.queue = sync_queue
        self.project_uri = project_uri

        self.daemon = True

    def run(self) -> None:
        ...

    def close(self) -> None:
        ...



import typing as t
from pathlib import Path

from starwhale.consts import RECOVER_DIRNAME, VERSION_PREFIX_CNT, DEFAULT_MANIFEST_NAME
from starwhale.base.uri import URI
from starwhale.utils.fs import guess_real_path
from starwhale.base.type import URIType
from starwhale.base.store import BaseStorage
from starwhale.utils.config import SWCliConfigMixed


class EvaluationStorage(BaseStorage):
    @property
    def recover_loc(self) -> Path:
        return (
            self.project_dir
            / URIType.EVALUATION
            / RECOVER_DIRNAME
            / self.id[:VERSION_PREFIX_CNT]
            / self.id
        )

    def _guess(self) -> t.Tuple[Path, str]:
        name = self.uri.object.name
        _p, _v, _ok = guess_real_path(
            self.project_dir / URIType.EVALUATION / name[:VERSION_PREFIX_CNT], name
        )
        return _p, _v

    @property
    def uri_type(self) -> str:
        return URIType.EVALUATION

    @property
    def manifest_path(self) -> Path:
        return self.loc / DEFAULT_MANIFEST_NAME

    @staticmethod
    def iter_all_jobs(project_uri: URI) -> t.Generator[t.Tuple[Path, bool], None, None]:
        # TODO: tune SWCliConfigMixed
        sw = SWCliConfigMixed()
        _job_dir = sw.rootdir / project_uri.project / URIType.EVALUATION
        for _path in _job_dir.glob(f"**/**/{DEFAULT_MANIFEST_NAME}"):
            yield _path, RECOVER_DIRNAME in _path.parts

    @staticmethod
    def local_run_dir(project_name: str, version: str) -> Path:
        # TODO: tune SWCliConfigMixed
        sw = SWCliConfigMixed()
        return (
            sw.rootdir
            / project_name
            / URIType.EVALUATION
            / version[:VERSION_PREFIX_CNT]
            / version
        )



from __future__ import annotations

import typing as t

import click
from textual.app import App
from textual.widgets import Header, ScrollView

from starwhale.base.uri import URI

from .widgets import Jobs, Models, Datasets, Runtimes
from .project_tree import ProjectTree, ProjectClick

if t.TYPE_CHECKING:
    from textual import events
    from textual.widget import Widget, RenderableType


class Dashboard(App):
    def __init__(self, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.main_table: t.Optional[RenderableType] = None
        self.project_uri = ""
        self.key_mapping: t.Dict[str, t.Callable] = {
            "M": Models,
            "D": Datasets,
            "R": Runtimes,
            "J": Jobs,
        }

    async def on_load(self, event: events.Load) -> None:
        pass

    async def on_mount(self, event: events.Mount) -> None:
        self.body = ScrollView(fluid=False)
        await self.view.dock(Header(style=""), edge="top")

        # pass empty dir to URI, it will use current instance / project
        uri = URI("")
        await self.view.dock(
            ScrollView(
                ProjectTree(
                    "Starwhale",
                    "projects",
                    current_instance=uri.instance_alias,
                    current_project=uri.project,
                )
            ),
            edge="left",
            size=30,
            name="sidebar",
        )
        await self.view.dock(self.body, edge="top")

    async def on_key(self, event: events.Key) -> None:
        if event.key in self.key_mapping:
            await self.update_main_table(self.key_mapping[event.key](self.project_uri))

    async def update_main_table(self, w: Widget) -> None:
        self.main_table = w
        await self.body.update(self.main_table)
        await self.main_table.focus()

    async def handle_project_click(self, message: ProjectClick) -> None:
        self.project_uri = message.uri
        await self.update_main_table(Models(self.project_uri))


@click.command(
    "board", help="A terminal based UI to interact with your starwhale instances"
)
def open_board() -> None:
    Dashboard.run(title="Starwhale")



from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod
from http import HTTPStatus
from collections import defaultdict

import yaml

from starwhale.utils import console, load_yaml
from starwhale.consts import (
    HTTPMethod,
    CREATED_AT_KEY,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import move_dir, empty_dir
from starwhale.base.type import URIType, BundleType, InstanceType
from starwhale.base.cloud import CloudRequestMixed, CloudBundleModelMixin
from starwhale.utils.http import ignore_error
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import NoSupportError
from starwhale.core.dataset.copy import DatasetCopy
from starwhale.api._impl.dataset.loader import DataRow

from .type import DatasetConfig, DatasetSummary
from .store import DatasetStorage

if t.TYPE_CHECKING:
    from starwhale.api._impl.dataset.model import Dataset as SDKDataset


class Dataset(BaseBundle, metaclass=ABCMeta):
    def __init__(self, uri: URI) -> None:
        self.store = DatasetStorage(uri)
        super().__init__(uri)

    def __str__(self) -> str:
        return f"Starwhale Dataset: {self.uri}"

    @abstractmethod
    def summary(self) -> t.Optional[DatasetSummary]:
        raise NotImplementedError

    def head(
        self, rows: int = 5, show_raw_data: bool = False
    ) -> t.List[t.Dict[str, t.Any]]:
        from starwhale.api._impl.dataset.model import Dataset as SDKDataset

        ret = []
        sds = SDKDataset.dataset(self.uri, readonly=True)
        for idx, row in enumerate(sds.head(n=rows, skip_fetch_data=not show_raw_data)):
            ret.append(
                {
                    "index": row.index,
                    "features": row.features,
                    "id": idx,
                }
            )

        return ret

    def diff(self, compare_uri: URI) -> t.Dict[str, t.Any]:
        # TODO: standalone or cloud dataset diff by datastore diff feature
        raise NotImplementedError

    @classmethod
    def get_dataset(cls, uri: URI) -> Dataset:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        dest_local_project_uri: str = "",
    ) -> None:
        dc = DatasetCopy(
            src_uri,
            dest_uri,
            URIType.DATASET,
            dest_local_project_uri=dest_local_project_uri,
        )
        dc.do()

    @classmethod
    def _get_cls(  # type: ignore
        cls, uri: URI
    ) -> t.Union[t.Type[StandaloneDataset], t.Type[CloudDataset]]:
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneDataset
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudDataset
        else:
            raise NoSupportError(f"dataset uri:{uri}")


class StandaloneDataset(Dataset, LocalStorageBundleMixin):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[
            str, t.Any
        ] = {}  # TODO: use manifest class get_conda_env
        self.yaml_name = DefaultYAMLName.DATASET
        self._version = uri.object.version

    def list_tags(self) -> t.List[str]:
        return self.tag.list()

    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.add(tags, ignore_errors)

    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.remove(tags, ignore_errors)

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.List[t.Dict[str, t.Any]]:
        _r = []

        for _bf in self.store.iter_bundle_history():
            _manifest_path = _bf.path / DEFAULT_MANIFEST_NAME
            if not _manifest_path.exists():
                continue

            _manifest = load_yaml(_bf.path / DEFAULT_MANIFEST_NAME)
            _r.append(
                dict(
                    name=self.name,
                    version=_bf.version,
                    size=_manifest.get("blobs_byte_size", 0),
                    created_at=_manifest[CREATED_AT_KEY],
                    tags=_bf.tags,
                    path=_bf.path,
                )
            )

        return _r

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: remove by tag
        if force:
            empty_dir(self.store.snapshot_workdir)
            return True, ""
        else:
            return move_dir(self.store.snapshot_workdir, self.store.recover_loc, False)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        dest_path = (
            self.store.bundle_dir / f"{self.uri.object.version}{BundleType.DATASET}"
        )
        return move_dir(self.store.recover_loc, dest_path, force)

    def info(self) -> t.Dict[str, t.Any]:
        return self._get_bundle_info()

    def summary(self) -> t.Optional[DatasetSummary]:
        _manifest = self.store.manifest
        _summary = _manifest.get("dataset_summary", {})
        return DatasetSummary(**_summary) if _summary else None

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filters = filters or {}
        rs = defaultdict(list)

        for _bf in DatasetStorage.iter_all_bundles(
            project_uri,
            bundle_type=BundleType.DATASET,
            uri_type=URIType.DATASET,
        ):
            if not cls.do_bundle_filter(_bf, filters):
                continue

            _mf = _bf.path / DEFAULT_MANIFEST_NAME
            if not _mf.exists():
                continue

            _manifest = load_yaml(_bf.path / DEFAULT_MANIFEST_NAME)

            rs[_bf.name].append(
                dict(
                    name=_bf.name,
                    version=_bf.version,
                    size=_manifest.get("dataset_summary", {}).get("blobs_byte_size", 0),
                    created_at=_manifest[CREATED_AT_KEY],
                    is_removed=_bf.is_removed,
                    path=_bf.path,
                    tags=_bf.tags,
                )
            )

        return rs, {}

    def build(self, *args: t.Any, **kwargs: t.Any) -> None:
        from starwhale.api._impl.dataset.model import Dataset as SDKDataset

        dataset_config: DatasetConfig = kwargs["config"]
        with SDKDataset.dataset(self.uri) as sds:
            sds = sds.with_builder_blob_config(
                volume_size=dataset_config.attr.volume_size,
                alignment_size=dataset_config.attr.alignment_size,
            )
            console.print(
                f":new: pending commit version: {sds.pending_commit_version[:SHORT_VERSION_CNT]}"
            )
            self._build_from_iterable_handler(sds, dataset_config)
            version = sds.commit()
            self._version = self.uri.object.version = version

            console.print(
                ":hibiscus: congratulation! you can run "
                f"[red bold blink] swcli dataset info {self.name}/version/{version[:SHORT_VERSION_CNT]}[/]"
            )

    def _build_from_iterable_handler(
        self, sds: SDKDataset, dataset_config: DatasetConfig
    ) -> None:
        _handler_name = getattr(dataset_config.handler, "__name__", None) or str(
            dataset_config.handler
        )

        console.print(
            f":ghost: iterate records from python iterable handler: {_handler_name}"
        )
        total = 0
        for index, item in enumerate(dataset_config.handler()):
            # TODO: support `--append`, `--update` and `--overwrite` for dataset build from iterable handler.
            # current build only supports `--update` feature.
            if isinstance(item, DataRow):
                key = item.index
            elif isinstance(item, (tuple, list)) and len(item) == 2:
                key = item[0]
            else:
                key = index  # auto-incr index
            sds[key] = item
            total += 1

        console.print(f":butterfly: update {total} records into {self.name} dataset")


class CloudDataset(CloudBundleModelMixin, Dataset):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filter_dict: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filter_dict = filter_dict or {}
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(
            project_uri, URIType.DATASET, page, size, filter_dict
        )

    def summary(self) -> t.Optional[DatasetSummary]:
        resp = self.do_http_request(
            f"/project/{self.uri.project}/{self.uri.object.typ}/{self.uri.object.name}",
            method=HTTPMethod.GET,
            instance_uri=self.uri,
            params={"versionUrl": self.uri.object.version},
            ignore_status_codes=[
                HTTPStatus.NOT_FOUND,
                HTTPStatus.INTERNAL_SERVER_ERROR,
            ],
        )
        if resp.status_code != HTTPStatus.OK:
            return None

        content = resp.json()
        _manifest: t.Dict[str, t.Any] = yaml.safe_load(
            content["data"].get("versionMeta", {})
        )
        _summary = _manifest.get("dataset_summary", {})
        return DatasetSummary(**_summary) if _summary else None

    def build(self, *args: t.Any, **kwargs: t.Any) -> None:
        raise NoSupportError("no support build dataset in the cloud instance")



from __future__ import annotations

import shutil
import typing as t
import fnmatch
from abc import ABC
from pathlib import Path

from starwhale.utils.config import SWCliConfigMixed
from starwhale.base.blob.file import PathLike, BlakeFile, OptionalPathLike

BuiltinPyExcludes = ["__pycache__/", "*.py[cod]", "*$py.class"]


class ObjectStore(ABC):
    def get(self, key: str) -> t.Any:
        raise NotImplementedError

    def put(self, path: PathLike, key: t.Optional[str]) -> t.Any:
        raise NotImplementedError


# TODO: support file lock for concurrent access
class LocalFileStore(ObjectStore):
    def __init__(self, root: OptionalPathLike = None, soft_link: bool = False) -> None:
        if root is None:
            root = SWCliConfigMixed().object_store_dir
        self.root = Path(root)
        self.root.mkdir(parents=True, exist_ok=True)
        self.soft_link = soft_link

    def get(self, key: str) -> t.Union[BlakeFile, None]:
        """
        Get a file from the object store.
        Args:
            key: the hash of the file
        Returns: the BlakeFile object got or None if not found
        """
        if not key:
            raise ValueError("key cannot be empty")
        path = Path(self.root) / key[:2] / key
        if not path.exists():
            return None
        return BlakeFile(path, key)

    def put(self, path: PathLike, key: t.Optional[str] = None) -> BlakeFile:
        """
        Copy a file to the object store.
        Args:
            path: the file to be copied
            key: the hash of the file, if not provided, will be calculated
        Returns: the BlakeFile object putted
        """
        return self._put(path, key)

    def link(self, src: PathLike, key: t.Optional[str] = None) -> None:
        """
        Link a file to the object store. only support hard link for now.
        Args:
            src: the file links from
            key: the hash of the file, if not provided, will be calculated
        """
        file = BlakeFile(src, key)
        dst = Path(self.root) / file.hash[:2] / file.hash
        if dst.exists():
            return
        file.link(dst)

    def _put(self, path: PathLike, key: t.Optional[str] = None) -> BlakeFile:
        file = BlakeFile(path, key)
        dst = Path(self.root) / file.hash[:2] / file.hash
        if not dst.exists():
            dst.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(path, str(dst))
        return BlakeFile(dst, file.hash)

    def copy_dir(
        self,
        src_dir: PathLike,
        dst_dir: PathLike,
        excludes: t.Optional[t.List[str]] = None,
    ) -> None:
        """
        Copy a directory from src_dir to the dst_dir using the object store cache.
        When the file is already in the object store, it will be linked instead of copied.
        Args:
            src_dir: the source directory
            dst_dir: the destination directory
            excludes: the files to be excluded, default to BuiltinPyExcludes (python cache files)
        """
        src_dir = Path(src_dir)
        dst_dir = Path(dst_dir)
        excludes = excludes or BuiltinPyExcludes
        for src in src_dir.rglob("*"):
            exclude = any(
                [fnmatch.fnmatch(str(src), str(src_dir / ex)) for ex in excludes]
            )
            if not src.is_file() or exclude:
                continue
            file = self._put(path=src)
            dst = dst_dir / src.relative_to(src_dir)
            dst.parent.mkdir(parents=True, exist_ok=True)
            file.link(dst, soft=self.soft_link)



import os
import typing as t

from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn

from starwhale.utils import console
from starwhale.consts import ENV_DISABLE_PROGRESS_BAR


def run_with_progress_bar(
    title: str,
    operations: t.Sequence[t.Tuple[t.Any, ...]],
    **kw: t.Any,
) -> None:
    if os.environ.get(ENV_DISABLE_PROGRESS_BAR) == "0":
        for op in operations:
            if len(op) == 4:
                op[0](**op[3])
            else:
                op[0]()
    else:
        with Progress(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            console=console,
            refresh_per_second=1,
        ) as progress:
            task = progress.add_task(
                f"[red]{title}", total=sum([o[1] for o in operations])
            )

            for idx, op in enumerate(operations):
                progress.update(task, description=f"[red]{op[2]}...")
                if len(op) == 4:
                    op[0](**op[3])
                else:
                    op[0]()
                progress.update(
                    task,
                    advance=op[1],
                    description=f"[green]{idx+1} out of {len(operations)} steps finished",
                )



from fastapi import APIRouter

from starwhale.version import STARWHALE_VERSION
from starwhale.web.response import success, SuccessResp

router = APIRouter()
prefix = "system"


@router.get("/version")
def version() -> SuccessResp:
    return success(STARWHALE_VERSION)



from __future__ import annotations

import os
import json
import typing as t
import urllib
import threading
from copy import deepcopy
from enum import Enum, unique
from queue import Queue
from types import TracebackType
from functools import partial
from collections import UserDict, defaultdict

import requests
from loguru import logger
from typing_extensions import Protocol

from starwhale.utils import validate_obj_name
from starwhale.consts import ENV_POD_NAME, STANDALONE_INSTANCE
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.base.mixin import ASDictMixin, _do_asdict_convert
from starwhale.consts.env import SWEnv
from starwhale.utils.error import (
    FormatError,
    NoSupportError,
    InvalidObjectName,
    FieldTypeOrValueError,
)
from starwhale.utils.retry import http_retry
from starwhale.utils.config import SWCliConfigMixed
from starwhale.api._impl.wrapper import Dataset as DatastoreWrapperDataset
from starwhale.api._impl.wrapper import DatasetTableKind
from starwhale.core.dataset.type import Link, JsonDict, BaseArtifact
from starwhale.api._impl.data_store import TableEmptyException

DEFAULT_CONSUMPTION_BATCH_SIZE = 50

_DR_DATA_KEY = "data_datastore_revision"
_DR_INFO_KEY = "info_datastore_revision"


class DatastoreRevision(t.NamedTuple):
    data: str
    info: str

    def asdict(self) -> t.Dict:
        return {
            _DR_DATA_KEY: self.data,
            _DR_INFO_KEY: self.info,
        }

    @classmethod
    def from_manifest(cls, manifest: t.Dict | None) -> DatastoreRevision:
        manifest = manifest or {}
        return DatastoreRevision(
            data=manifest.get(_DR_DATA_KEY, ""),
            info=manifest.get(_DR_INFO_KEY, ""),
        )


class TabularDatasetInfo(UserDict):
    _ROW_ID = 0

    def __init__(self, mapping: t.Any = None, **kwargs: t.Any) -> None:
        mapping = mapping or {}
        if not isinstance(mapping, dict):
            raise TypeError(f"data is not dict type: {mapping}")

        if kwargs:
            mapping.update(kwargs)

        converted_mapping: t.Dict[str, t.Any] = {}
        for k, v in mapping.items():
            if not isinstance(k, str):
                raise TypeError(f"key:{k} is not str type")

            # TODO： add validator for value?
            converted_mapping[k] = JsonDict.from_data(v)
        super().__init__(converted_mapping)

    def __getitem__(self, k: str) -> t.Any:
        return JsonDict.to_data(super().__getitem__(k))

    def __setitem__(self, k: str, v: t.Any) -> None:
        if not isinstance(k, str):
            raise TypeError(f"key:{k} is not str type")
        super().__setitem__(k, JsonDict.from_data(v))

    @classmethod
    def load_from_datastore(
        cls, ds_wrapper: DatastoreWrapperDataset
    ) -> TabularDatasetInfo:
        try:
            rows = list(ds_wrapper.scan(cls._ROW_ID, cls._ROW_ID, end_inclusive=True))
            rows_cnt = len(rows)
            if rows_cnt == 1:
                d = rows[0]
            elif rows_cnt > 1:
                raise RuntimeError(f"fetch multi info rows: {rows}")
            else:
                d = {}
        except TableEmptyException:
            d = {}

        d.pop("id", None)
        return cls(d)

    def save_to_datastore(self, ds_wrapper: DatastoreWrapperDataset) -> str:
        ds_wrapper.put(data_id=self._ROW_ID, **self.data)
        return ds_wrapper.flush()


class TabularDatasetRow(ASDictMixin):

    _FEATURES_PREFIX = "features/"

    def __init__(
        self,
        id: t.Union[str, int],
        features: t.Optional[t.Dict[str, t.Any]] = None,
        **kw: t.Union[str, int, float],
    ) -> None:
        self.id = id
        self.features = features or {}
        self.extra_kw = kw
        # TODO: add non-starwhale object store related fields, such as address, authority
        # TODO: add data uri crc for versioning
        self._do_validate()

    @classmethod
    def from_datastore(
        cls,
        id: t.Union[str, int],
        **kw: t.Any,
    ) -> TabularDatasetRow:
        _content = {}
        _extra_kw = {}
        for k, v in kw.items():
            if k.startswith(cls._FEATURES_PREFIX):
                _, name = k.split(cls._FEATURES_PREFIX, 1)
                _content[name] = JsonDict.to_data(v)
            else:
                _extra_kw[k] = v

        return cls(
            id=id,
            features=_content,
            **_extra_kw,
        )

    def __eq__(self, o: object) -> bool:
        s = deepcopy(self.__dict__)
        o = deepcopy(o.__dict__)
        return s == o

    def _do_validate(self) -> None:
        if not isinstance(self.id, (str, int)):
            raise FieldTypeOrValueError(f"id is not int or str type: {self.id}")

        if self.id == "":
            raise FieldTypeOrValueError("id is empty")

        if not isinstance(self.features, dict) or not self.features:
            raise FieldTypeOrValueError("no data field")

    def __str__(self) -> str:
        return f"row-{self.id}"

    def __repr__(self) -> str:
        return f"row-{self.id}" f"{self.features} "

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        d = super().asdict(ignore_keys=ignore_keys or ["features", "extra_kw"])
        d.update(_do_asdict_convert(self.extra_kw))
        for k, v in self.features.items():
            d[f"{self._FEATURES_PREFIX}{k}"] = JsonDict.from_data(v)
        return d

    @classmethod
    def artifacts_of(cls, features: t.Dict) -> t.List[BaseArtifact]:
        artifacts = []
        for v in features.values():
            if isinstance(v, dict):
                artifacts.extend(cls.artifacts_of(v))
            elif isinstance(v, BaseArtifact):
                artifacts.append(v)
        return artifacts

    @property
    def artifacts(self) -> t.List[BaseArtifact]:
        return TabularDatasetRow.artifacts_of(self.features)


_TDType = t.TypeVar("_TDType", bound="TabularDataset")


class TabularDataset:
    def __init__(
        self,
        name: str,
        project: str,
        start: t.Optional[t.Any] = None,
        end: t.Optional[t.Any] = None,
        instance_name: str = "",
        token: str = "",
        data_datastore_revision: str = "",
        info_datastore_revision: str = "",
    ) -> None:
        _ok, _reason = validate_obj_name(name)
        if not _ok:
            raise InvalidObjectName(f"{name}: {_reason}")
        self.name = name

        self.project = project
        self.instance_name = instance_name

        dwd = partial(
            DatastoreWrapperDataset,
            dataset_name=name,
            project=project,
            instance_name=instance_name,
            token=token,
        )
        self._ds_wrapper = dwd(
            kind=DatasetTableKind.META, dataset_scan_revision=data_datastore_revision
        )
        self._info_ds_wrapper = dwd(
            kind=DatasetTableKind.INFO, dataset_scan_revision=info_datastore_revision
        )

        self._info_changed = False
        self._info: t.Optional[TabularDatasetInfo] = None
        self._info_lock = threading.Lock()

        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"Dataset Table: {self._ds_wrapper}"

    __repr__ = __str__

    def update(
        self, row_id: t.Union[str, int], **kw: t.Union[int, str, bytes, Link]
    ) -> None:
        self._ds_wrapper.put(row_id, **kw)

    def put(self, row: TabularDatasetRow) -> None:
        self._ds_wrapper.put(row.id, **row.asdict())

    def delete(self, row_id: t.Union[str, int]) -> None:
        self._ds_wrapper.delete(row_id)

    def flush(self) -> t.Tuple[str, str]:
        info_revision = ""
        if self._info is not None and self._info_changed:
            info_revision = self._info.save_to_datastore(self._info_ds_wrapper)

        data_revision = self._ds_wrapper.flush()
        return data_revision, info_revision

    def scan(
        self,
        start: t.Optional[t.Any] = None,
        end: t.Optional[t.Any] = None,
        end_inclusive: bool = False,
        revision: str = "",
    ) -> t.Generator[TabularDatasetRow, None, None]:
        if start is None or (self.start is not None and start < self.start):
            start = self.start

        if end is None or (self.end is not None and end > self.end):
            end = self.end

        for _d in self._ds_wrapper.scan(start, end, end_inclusive, revision=revision):
            yield TabularDatasetRow.from_datastore(**_d)

    def scan_batch(
        self,
        start: t.Optional[t.Any] = None,
        end: t.Optional[t.Any] = None,
        batch_size: int = 32,
        revision: str = "",
    ) -> t.Generator[t.List[TabularDatasetRow], None, None]:
        batch = []
        for r in self.scan(start, end, revision=revision):
            batch.append(r)
            if len(batch) % batch_size == 0:
                yield batch
                batch = []
        if batch:
            yield batch

    def close(self) -> None:
        self.flush()
        self._ds_wrapper.close()
        self._info_ds_wrapper.close()

    def __enter__(self: _TDType) -> _TDType:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            logger.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self.close()

    @classmethod
    def from_uri(
        cls: t.Type[_TDType],
        uri: URI,
        start: t.Optional[t.Any] = None,
        end: t.Optional[t.Any] = None,
        data_datastore_revision: str = "",
        info_datastore_revision: str = "",
    ) -> _TDType:
        return cls(
            name=uri.object.name,
            project=uri.project,
            start=start,
            end=end,
            instance_name=uri.instance,
            data_datastore_revision=data_datastore_revision,
            info_datastore_revision=info_datastore_revision,
        )

    @property
    def info(self) -> TabularDatasetInfo:
        if self._info is not None:
            return self._info

        with self._info_lock:
            self._info = TabularDatasetInfo.load_from_datastore(self._info_ds_wrapper)
        return self._info

    @info.setter
    def info(self, data: t.Optional[t.Dict[str, t.Any]]) -> None:
        if data is None:
            return

        if not isinstance(data, dict):
            raise TypeError(
                f"data:{type(data)} {data} is not dict type for info update"
            )

        with self._info_lock:
            self._info_changed = True
            self._info = TabularDatasetInfo(data)


@unique
class _RunEnvType(Enum):
    POD = "pod"
    THREAD = "thread"
    PROCESS = "process"


class TabularDatasetSessionConsumption(Protocol):
    batch_size: int

    def get_scan_range(
        self, processed_keys: t.Optional[t.List[t.Tuple[t.Any, t.Any]]] = None
    ) -> t.Optional[t.Tuple[t.Any, t.Any]]:
        ...  # pragma: no cover


local_standalone_tdsc: t.Dict[str, StandaloneTDSC] = {}
lock_s_tdsc = threading.Lock()


def get_dataset_consumption(
    dataset_uri: t.Union[str, URI],
    session_id: str,
    batch_size: t.Optional[int] = None,
    session_start: t.Optional[t.Any] = None,
    session_end: t.Optional[t.Any] = None,
    instance_uri: str = "",
    instance_token: str = "",
) -> TabularDatasetSessionConsumption:
    # TODO: tune factory class arguments
    _uri = instance_uri or os.environ.get(SWEnv.instance_uri)

    if isinstance(dataset_uri, str):
        dataset_uri = URI(dataset_uri, expected_type=URIType.DATASET)

    batch_size = batch_size or int(
        os.environ.get(
            "DATASET_CONSUMPTION_BATCH_SIZE",
            DEFAULT_CONSUMPTION_BATCH_SIZE,
        )
    )
    if _uri is None or _uri == STANDALONE_INSTANCE:
        global local_standalone_tdsc
        key = f"{dataset_uri}-{session_id}-{session_start}-{session_end}-{batch_size}"
        with lock_s_tdsc:
            _obj = local_standalone_tdsc.get(key)
            if not _obj:
                _obj = StandaloneTDSC(
                    dataset_uri=dataset_uri,
                    session_id=session_id,
                    batch_size=batch_size,
                    session_start=session_start,
                    session_end=session_end,
                )
                local_standalone_tdsc[key] = _obj
            return _obj
    else:
        _token = (
            instance_token
            or SWCliConfigMixed().get_sw_token(instance=_uri)
            or os.getenv(SWEnv.instance_token, "")
        )
        return CloudTDSC(
            instance_uri=_uri,
            dataset_uri=dataset_uri,
            session_id=session_id,
            batch_size=batch_size,
            session_start=session_start,
            session_end=session_end,
            instance_token=_token,
        )


class StandaloneTDSC(TabularDatasetSessionConsumption):
    class _BatchTask:
        def __init__(self, start: t.Any, end: t.Any) -> None:
            self.start = start
            self.end = end

    def __init__(
        self,
        dataset_uri: URI,
        session_id: str,
        batch_size: int = DEFAULT_CONSUMPTION_BATCH_SIZE,
        session_start: t.Optional[t.Any] = None,
        session_end: t.Optional[t.Any] = None,
    ) -> None:
        if dataset_uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError(
                f"StandaloneTDSC only supports standalone dataset: {dataset_uri}"
            )

        self.project = dataset_uri.project
        self.dataset_name = dataset_uri.object.name
        self.dataset_version = dataset_uri.object.version

        self.session_id = session_id
        if batch_size <= 0:
            raise FieldTypeOrValueError(f"batch_size is invalid: {batch_size}")
        self.batch_size = batch_size
        self.session_start = session_start
        self.session_end = session_end
        self.run_env = _RunEnvType.THREAD

        self._lock = threading.Lock()
        self._todo_queue = self._init_dataset_todo_queue()
        self._doing_consumption: t.Dict[str, t.Dict[str, t.Any]] = defaultdict(dict)

        # TODO: support sidecar thread monitor
        # TODO: support max_retries

    def _init_dataset_todo_queue(self) -> Queue:
        # TODO: support datastore revision
        wrapper = DatastoreWrapperDataset(
            dataset_name=self.dataset_name,
            project=self.project,
        )
        ids = [i["id"] for i in wrapper.scan_id(self.session_start, self.session_end)]
        id_cnt = len(ids)
        queue: Queue[StandaloneTDSC._BatchTask] = Queue(maxsize=id_cnt)
        for start in range(0, id_cnt, self.batch_size):
            end = start + self.batch_size
            # TODO: add put block?
            start_key = ids[start]
            end_key = ids[end] if end < id_cnt else None
            queue.put(StandaloneTDSC._BatchTask(start=start_key, end=end_key))
        return queue

    def get_scan_range(
        self,
        processed_keys: t.Optional[t.List[t.Tuple[t.Any, t.Any]]] = None,
    ) -> t.Optional[t.Tuple[t.Any, t.Any]]:
        processed_keys = processed_keys or []
        consumer_id = self.consumer_id

        with self._lock:
            consumer: t.Dict = self._doing_consumption.get(consumer_id, {})
            if consumer:
                for k in processed_keys:
                    if not k:
                        continue
                    consumer.pop(f"{k[0]}-{k[1]}", None)

            if not self._todo_queue.empty():
                task: StandaloneTDSC._BatchTask = self._todo_queue.get()
                self._doing_consumption[consumer_id][f"{task.start}-{task.end}"] = task
                logger.info(
                    f"{consumer_id} handle scan-range: ({task.start}, {task.end})"
                )
                return task.start, task.end

            for cid in list(self._doing_consumption.keys()):
                if len(self._doing_consumption[cid]) == 0:
                    self._doing_consumption.pop(cid, None)

        return None

    def __str__(self) -> str:
        return f"[Standalone]Dataset Consumption: id-{self.session_id}, range-[{self.session_start},{self.session_end}], consumer:{self.consumer_id}@{self.run_env}"

    def __repr__(self) -> str:
        return f"[Standalone]Dataset Consumption: id-{self.session_id}, range-[{self.session_start},{self.session_end}], consumer:{self.consumer_id}@{self.run_env}, batch:{self.batch_size}"

    @property
    def consumer_id(self) -> str:
        if self.run_env == _RunEnvType.THREAD:
            return f"thread-{id(threading.current_thread())}"
        else:
            return f"process-{os.getpid()}"  # pragma: no cover


class CloudTDSC(TabularDatasetSessionConsumption):
    def __init__(
        self,
        instance_uri: str,
        dataset_uri: URI,
        session_id: str,
        batch_size: int = DEFAULT_CONSUMPTION_BATCH_SIZE,
        session_start: t.Optional[t.Any] = None,
        session_end: t.Optional[t.Any] = None,
        instance_token: str = "",
    ) -> None:
        self.instance_uri = instance_uri
        self.instance_token = instance_token
        self.session_id = session_id
        self.batch_size = batch_size
        self.session_start = session_start
        self.session_end = session_end
        self.dataset_uri = dataset_uri
        self.run_env = _RunEnvType.POD
        self.consumer_id = os.environ.get(ENV_POD_NAME)

        self._do_validate()

    def _do_validate(self) -> None:
        if not self.instance_token:
            raise FieldTypeOrValueError("instance token is empty")

        if (
            not self.dataset_uri.project
            or not self.dataset_uri.object.name
            or not self.dataset_uri.object.version
        ):
            raise FormatError(f"wrong dataset uri format: {self.dataset_uri}")

        if not self.consumer_id:
            raise RuntimeError("failed to get pod name")

    @http_retry
    def get_scan_range(
        self, processed_keys: t.Optional[t.List[t.Tuple[t.Any, t.Any]]] = None
    ) -> t.Optional[t.Tuple[t.Any, t.Any]]:
        post_data = {
            "batchSize": self.batch_size,
            "sessionId": self.session_id,
            "consumerId": self.consumer_id,
        }
        if processed_keys is not None:
            post_data["processedData"] = [
                {"start": p[0], "end": p[1]} for p in processed_keys
            ]

        if self.session_start is not None:
            post_data["start"] = self.session_start

        if self.session_end is not None:
            post_data["end"] = self.session_end

        resp = requests.post(
            urllib.parse.urljoin(
                self.instance_uri,
                f"api/v1/project/{self.dataset_uri.project}/dataset/{self.dataset_uri.object.name}/version/{self.dataset_uri.object.version}/consume",
            ),
            data=json.dumps(post_data),
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Authorization": self.instance_token,  # type: ignore
            },
            timeout=300,
        )
        resp.raise_for_status()
        range_data = resp.json()["data"]

        return None if range_data is None else (range_data["start"], range_data["end"])

    def __str__(self) -> str:
        return (
            f"[Cloud:{self.instance_uri}]Dataset Consumption: id-{self.session_id}, "
            f"range-[{self.session_start},{self.session_end}], consumer:{self.consumer_id}@{self.run_env}"
        )

    def __repr__(self) -> str:
        return (
            f"[Cloud:{self.instance_uri}]Dataset Consumption: id-{self.session_id}, "
            f"range-[{self.session_start},{self.session_end}], consumer:{self.consumer_id}@{self.run_env}, "
            f"batch-{self.batch_size}, dataset-{self.dataset_uri}"
        )



from fastapi import APIRouter

from starwhale.web.response import success, SuccessResp

router = APIRouter()
prefix = "user"

user = {
    "id": "1",
    "name": "starwhale",
    "createdTime": 0,
    "isEnabled": True,
    "systemRole": "OWNER",
    "projectRoles": {"1": "OWNER"},
}


@router.get("/current")
def current() -> SuccessResp:
    return success(user)



from __future__ import annotations

import io
import os
import sys
import json
import shutil
import typing as t
import threading
from abc import ABCMeta, abstractmethod
from types import TracebackType
from pathlib import Path
from urllib.parse import urlparse

import boto3
import requests
from loguru import logger
from botocore.client import Config as S3Config
from typing_extensions import Protocol

from starwhale.consts import (
    HTTPMethod,
    SWDSBackendType,
    VERSION_PREFIX_CNT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import (
    ensure_dir,
    blake2b_file,
    FilePosition,
    BLAKE2B_SIGNATURE_ALGO,
)
from starwhale.base.type import URIType, BundleType, InstanceType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.base.store import BaseStorage
from starwhale.utils.error import (
    FormatError,
    NotFoundError,
    NoSupportError,
    InvalidObjectName,
    FieldTypeOrValueError,
)
from starwhale.utils.retry import http_retry
from starwhale.utils.config import SWCliConfigMixed, get_swcli_config_path
from starwhale.core.dataset.type import Link

# TODO: refactor Dataset and ModelPackage LocalStorage
_DEFAULT_S3_REGION = "local"
_DEFAULT_S3_ENDPOINT = "localhost:9000"
_DEFAULT_S3_BUCKET = "starwhale"

# TODO: config chunk size
_CHUNK_SIZE = 8 * 1024 * 1024  # 8MB


class DatasetStorage(BaseStorage):
    object_hash_algo = BLAKE2B_SIGNATURE_ALGO
    short_sign_cnt = 16

    def _guess(self) -> t.Tuple[Path, str]:
        return self._guess_for_bundle()

    @property
    def recover_loc(self) -> Path:
        return self._get_recover_loc_for_bundle()

    @property
    def snapshot_workdir(self) -> Path:
        if self.building:
            return self.tmp_dir
        version = self.uri.object.version
        return (
            self.project_dir
            / URIType.DATASET
            / self.uri.object.name
            / version[:VERSION_PREFIX_CNT]
            / f"{version}{BundleType.DATASET}"
        )

    @property
    def manifest_path(self) -> Path:
        return self.loc / DEFAULT_MANIFEST_NAME

    @property
    def bundle_type(self) -> str:
        return BundleType.DATASET

    @property
    def uri_type(self) -> str:
        return URIType.DATASET

    @property
    def data_dir(self) -> Path:
        return self.snapshot_workdir / "data"

    @property
    def src_dir(self) -> Path:
        return self.snapshot_workdir / "src"

    @property
    def dataset_rootdir(self) -> Path:
        return self.project_dir / URIType.DATASET

    @classmethod
    def save_data_file(
        cls, src: t.Union[Path, str], force: bool = False, remove_src: bool = False
    ) -> t.Tuple[str, Path]:
        src = Path(src)

        if not src.exists():
            raise NotFoundError(f"data origin file: {src}")

        if not src.is_file():
            raise NoSupportError(f"{src} is not file")

        sign_name = blake2b_file(src)
        dest = cls._get_object_store_path(sign_name)
        if dest.exists() and not force:
            return sign_name, dest

        ensure_dir(dest.parent)

        _src, _dest = str(src.absolute()), str(dest.absolute())
        if remove_src:
            shutil.move(_src, _dest)
        else:
            shutil.copyfile(_src, _dest)
        return sign_name, dest

    @classmethod
    def _get_object_store_path(cls, hash_name: str) -> Path:
        _prefix_cnt = 2
        hash_name = hash_name.strip()
        if len(hash_name) < _prefix_cnt:
            raise InvalidObjectName(f"hash name({hash_name}) is too short")

        return (
            SWCliConfigMixed().object_store_dir
            / cls.object_hash_algo
            / hash_name[:_prefix_cnt]
            / hash_name
        )

    def get_data_file(self, name: str) -> Path:
        path = self._get_object_store_path(name)
        if path.exists():
            return path

        return self.data_dir / name

    def get_all_data_files(self) -> t.List[Path]:
        files = []
        for f in self.data_dir.rglob("*"):
            if not f.is_symlink():
                continue
            files.append(f.resolve().absolute())
        return files


class S3Uri(t.NamedTuple):
    bucket: str
    key: str
    protocol: str = "s3"
    endpoint: str = _DEFAULT_S3_ENDPOINT


_TFLType = t.TypeVar("_TFLType", covariant=True)


class FileLikeObj(Protocol[_TFLType]):
    def __enter__(self) -> _TFLType:
        ...

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        ...

    def read(self, size: int) -> bytes:
        ...

    def close(self) -> None:
        ...


class S3Connection:
    connections_config: t.List[S3Connection] = []
    init_config_lock = threading.Lock()
    supported_schemes = {"s3", "minio", "aliyun", "oss"}
    DEFAULT_CONNECT_TIMEOUT = 10.0
    DEFAULT_READ_TIMEOUT = 50.0
    DEFAULT_MAX_ATTEMPTS = 6

    def __init__(
        self,
        endpoint: str,
        access_key: str,
        secret_key: str,
        region: str = "",
        bucket: str = "",
        connect_timeout: float = DEFAULT_CONNECT_TIMEOUT,
        read_timeout: float = DEFAULT_READ_TIMEOUT,
        total_max_attempts: int = DEFAULT_MAX_ATTEMPTS,
    ) -> None:
        self.endpoint = endpoint.strip()
        if self.endpoint and not self.endpoint.startswith(("http://", "https://")):
            self.endpoint = f"http://{self.endpoint}"
        r = urlparse(self.endpoint)
        self.endpoint_loc = r.netloc.split("@")[-1]
        self.access_key = access_key
        self.secret_key = secret_key
        self.region = region
        self.bucket = bucket
        self.connect_timeout = float(
            os.environ.get("SW_S3_CONNECT_TIMEOUT", connect_timeout)
        )
        self.read_timeout = float(os.environ.get("SW_S3_READ_TIMEOUT", read_timeout))
        self.total_max_attempts = int(
            os.environ.get("SW_S3_TOTAL_MAX_ATTEMPTS", total_max_attempts)
        )
        # configs like {"addressing_style": "auto"}
        # more info in: botocore.Config._validate_s3_configuration
        # https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
        self.extra_s3_configs = json.loads(os.environ.get("SW_S3_EXTRA_CONFIGS", "{}"))

    def fits(self, bucket: str, endpoint_loc: str) -> bool:
        return self.bucket == bucket and self.endpoint_loc == endpoint_loc

    def __str__(self) -> str:
        return f"endpoint[{self.endpoint}]-region[{self.region}]"

    __repr__ = __str__

    @classmethod
    def from_uri(cls, uri: str) -> S3Connection:
        """make S3 Connection by uri

        uri:
            - s3://username:password@127.0.0.1:8000/bucket/key
            - s3://127.0.0.1:8000/bucket/key
        """
        r = urlparse(uri.strip())
        if r.scheme not in cls.supported_schemes:
            raise NoSupportError(
                f"s3 connection only support {cls.supported_schemes} prefix, the actual uri is {uri}"
            )
        endpoint_loc = r.netloc.split("@")[-1]
        parts = r.path.lstrip("/").split("/", 1)
        if len(parts) != 2 or parts[0] == "" or parts[1] == "":
            raise FieldTypeOrValueError(
                f"{uri} is not a valid s3 uri for bucket and key"
            )
        bucket = parts[0]
        cls.build_init_connections()
        for connection in S3Connection.connections_config:
            if connection.fits(bucket, endpoint_loc):
                return connection
        if r.username and r.password:
            return cls(
                endpoint=endpoint_loc,
                access_key=r.username,
                secret_key=r.password,
                region=_DEFAULT_S3_REGION,
                bucket=bucket,
            )
        raise NoSupportError(f"no matching s3 config in {get_swcli_config_path()}")

    @classmethod
    def build_init_connections(cls) -> None:
        with S3Connection.init_config_lock:
            if S3Connection.connections_config:
                return
            sw_config = SWCliConfigMixed()
            link_auths = sw_config.link_auths
            if not link_auths:
                return
            for la in link_auths:
                if not la.get("type") or type(la.get("type")) != str:
                    continue
                if la.get("type") not in cls.supported_schemes:
                    continue
                if (
                    not la.get("endpoint")
                    or not la.get("ak")
                    or not la.get("sk")
                    or not la.get("bucket")
                ):
                    continue
                S3Connection.connections_config.append(
                    cls(
                        endpoint=la.get("endpoint"),
                        access_key=la.get("ak"),
                        secret_key=la.get("sk"),
                        region=la.get("region") or _DEFAULT_S3_REGION,
                        bucket=la.get("bucket"),
                        connect_timeout=la.get(
                            "connect_timeout", cls.DEFAULT_CONNECT_TIMEOUT
                        ),
                        read_timeout=la.get("read_timeout", cls.DEFAULT_READ_TIMEOUT),
                        total_max_attempts=la.get(
                            "total_max_attempts", cls.DEFAULT_MAX_ATTEMPTS
                        ),
                    )
                )
            env_conn = cls.from_env()
            if env_conn:
                S3Connection.connections_config.append(env_conn)

    @classmethod
    def from_env(cls) -> S3Connection:
        # TODO: support multi s3 backend servers
        _env = os.environ.get
        return S3Connection(
            endpoint=_env("SW_S3_ENDPOINT", _DEFAULT_S3_ENDPOINT),
            access_key=_env("SW_S3_ACCESS_KEY", ""),
            secret_key=_env("SW_S3_SECRET", ""),
            region=_env("SW_S3_REGION", _DEFAULT_S3_REGION),
            bucket=_env("SW_S3_BUCKET", _DEFAULT_S3_BUCKET),
        )


class ObjectStore:
    _store_lock = threading.Lock()
    _stores: t.Dict[str, ObjectStore] = {}

    def __init__(
        self,
        backend: str,
        bucket: str = "",
        dataset_uri: URI = URI(""),
        key_prefix: str = "",
        **kw: t.Any,
    ) -> None:
        self.bucket = bucket

        self.backend: StorageBackend
        if backend == SWDSBackendType.S3:
            conn = kw.get("conn") or S3Connection.from_env()
            self.backend = S3StorageBackend(conn)
        elif backend == SWDSBackendType.SignedUrl:
            self.backend = SignedUrlBackend(dataset_uri)
        elif backend == SWDSBackendType.Http:
            self.backend = HttpBackend()
        else:
            self.backend = LocalFSStorageBackend()

        self.key_prefix = key_prefix or os.environ.get("SW_OBJECT_STORE_KEY_PREFIX", "")

    def __str__(self) -> str:
        return f"ObjectStore backend:{self.backend}"

    def __repr__(self) -> str:
        return f"ObjectStored:{self.backend}, bucket:{self.bucket}, key_prefix:{self.key_prefix}"

    @classmethod
    def from_data_link_uri(cls, data_link: Link) -> ObjectStore:
        if not data_link:
            raise FieldTypeOrValueError("data_link is empty")

        # TODO: support other uri type
        if data_link.scheme in S3Connection.supported_schemes:
            backend = SWDSBackendType.S3
            conn = S3Connection.from_uri(data_link.uri)
            bucket = conn.bucket
        elif data_link.scheme in ["http", "https"]:
            backend = SWDSBackendType.Http
            bucket = ""
            conn = None
        else:
            backend = SWDSBackendType.LocalFS
            bucket = ""
            conn = None

        return cls(backend=backend, bucket=bucket, conn=conn)

    @classmethod
    def from_dataset_uri(cls, dataset_uri: URI) -> ObjectStore:
        if dataset_uri.object.typ != URIType.DATASET:
            raise NoSupportError(f"{dataset_uri} is not dataset uri")
        return cls(
            backend=SWDSBackendType.LocalFS,
            bucket=str(DatasetStorage(dataset_uri).data_dir.absolute()),
        )

    @classmethod
    def to_signed_http_backend(cls, dataset_uri: URI) -> ObjectStore:
        if dataset_uri.object.typ != URIType.DATASET:
            raise NoSupportError(f"{dataset_uri} is not dataset uri")
        return cls(backend=SWDSBackendType.SignedUrl, dataset_uri=dataset_uri)

    @classmethod
    def get_store(
        cls, link: Link, owner: t.Optional[t.Union[str, URI]] = None
    ) -> ObjectStore:
        with cls._store_lock:
            _up = urlparse(link.uri)
            _parts = _up.path.lstrip("/").split("/", 1)
            _cache_key = link.uri.replace(_parts[-1], "")
            _k = f"{owner}.{_cache_key}"
            _store = cls._stores.get(_k)
            if _store:
                return _store
            if not owner:
                _store = ObjectStore.from_data_link_uri(link)
                cls._stores[_k] = _store
                return _store
            if isinstance(owner, str):
                owner = URI(owner, expected_type=URIType.DATASET)
            if owner.instance_type == InstanceType.CLOUD:
                _store = ObjectStore.to_signed_http_backend(owner)
            else:
                if link.scheme and link.scheme != "fs":
                    _store = ObjectStore.from_data_link_uri(link)
                else:
                    _store = ObjectStore.from_dataset_uri(owner)

            logger.debug(f"new store backend created for key: {_k}")
            cls._stores[_k] = _store
            return _store


class StorageBackend(metaclass=ABCMeta):
    def __init__(
        self,
        kind: str,
    ) -> None:
        self.kind = kind

    def __str__(self) -> str:
        return f"StorageBackend for {self.kind}"

    __repr__ = __str__

    @abstractmethod
    def _make_file(
        self, key_compose: t.Tuple[Link, int, int], **kw: t.Any
    ) -> FileLikeObj:
        raise NotImplementedError


class S3StorageBackend(StorageBackend):
    lock_s3_creation = threading.Lock()

    def __init__(
        self,
        conn: S3Connection,
    ):
        super().__init__(kind=SWDSBackendType.S3)
        with S3StorageBackend.lock_s3_creation:
            self.s3 = boto3.resource(
                "s3",
                endpoint_url=conn.endpoint,
                aws_access_key_id=conn.access_key,
                aws_secret_access_key=conn.secret_key,
                config=S3Config(
                    s3=conn.extra_s3_configs,
                    connect_timeout=conn.connect_timeout,
                    read_timeout=conn.read_timeout,
                    signature_version="s3v4",
                    retries={
                        "total_max_attempts": conn.total_max_attempts,
                        "mode": "standard",
                    },
                ),
                region_name=conn.region,
            )

    def _make_file(
        self, key_compose: t.Tuple[Link, int, int], **kw: t.Any
    ) -> FileLikeObj:
        # TODO: merge connections for s3
        _key, _start, _end = key_compose
        return S3BufferedFileLike(
            s3=self.s3,
            bucket=kw["bucket"],
            key=_key.uri,
            start=_start,
            end=_end,
        )


class LocalFSStorageBackend(StorageBackend):
    def __init__(self) -> None:
        super().__init__(kind=SWDSBackendType.LocalFS)

    def _make_file(
        self, key_compose: t.Tuple[Link, int, int], **kw: t.Any
    ) -> FileLikeObj:
        _key_l, _start, _end = key_compose
        _key = _key_l.uri
        # TODO: tune reopen file performance, merge files
        data_path = DatasetStorage._get_object_store_path(_key)
        # TODO: remove dataset data_dir
        if not data_path.exists():
            bucket = kw["bucket"]
            bucket_path = (
                Path(bucket).expanduser() if bucket.startswith("~/") else Path(bucket)
            )
            data_path = bucket_path / _key[: DatasetStorage.short_sign_cnt]
            if not data_path.exists():
                data_path = bucket_path / _key

        _end = min(_end, data_path.stat().st_size)
        with data_path.open("rb") as f:
            f.seek(_start)
            return io.BytesIO(f.read(_end - _start + 1))  # type: ignore


class SignedUrlBackend(StorageBackend, CloudRequestMixed):
    def __init__(self, dataset_uri: URI) -> None:
        super().__init__(kind=SWDSBackendType.SignedUrl)
        self.dataset_uri = dataset_uri

    @http_retry
    def _make_file(
        self, key_compose: t.Tuple[Link, int, int], **kw: t.Any
    ) -> FileLikeObj:
        _key, _start, _end = key_compose
        return HttpBufferedFileLike(
            url=_key.signed_uri or self.sign_uri(_key.uri),
            headers={"Range": f"bytes={_start or 0}-{_end or sys.maxsize}"},
            timeout=90,
        )

    def sign_uri(self, uri: str) -> str:
        r = self.do_http_request(
            f"/project/{self.dataset_uri.project}/{self.dataset_uri.object.typ}/{self.dataset_uri.object.name}/uri/sign-links",
            method=HTTPMethod.POST,
            instance_uri=self.dataset_uri,
            params={
                "expTimeMillis": int(
                    os.environ.get("SW_MODEL_PROCESS_UNIT_TIME_MILLIS", "60000")
                ),
            },
            json=[uri],
            use_raise=True,
        ).json()
        return r["data"].get(uri, "")  # type: ignore


class HttpBackend(StorageBackend, CloudRequestMixed):
    def __init__(self) -> None:
        super().__init__(kind=SWDSBackendType.Http)

    @http_retry
    def _make_file(
        self, key_compose: t.Tuple[Link, int, int], **kw: t.Any
    ) -> FileLikeObj:
        _key, _start, _end = key_compose
        return HttpBufferedFileLike(
            url=_key.uri,
            headers={"Range": f"bytes={_start or 0}-{_end or sys.maxsize}"},
            timeout=90,
        )


_BFType = t.TypeVar("_BFType", bound="BaseBufferedFileLike")


class BaseBufferedFileLike(metaclass=ABCMeta):
    def __init__(self, buffer_size: int) -> None:
        self._read_buffer = BytesBuffer(buffer_size)

    def __enter__(self: _BFType) -> _BFType:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            print(f"type:{type}, exception:{value}, traceback:{trace}")

        self.close()

    def read(self, size: int) -> bytes:
        ret = self._read(size)
        if isinstance(ret, memoryview):
            return ret.tobytes()
        elif isinstance(ret, bytes):
            return ret
        else:
            raise FormatError(f"{ret}:{type(ret)} is not bytes or memoryview type")

    def _read(self, size: int) -> t.Union[memoryview, bytes]:
        if size == 0:
            return b""
        elif size < 0:
            return self._read_buffer.read() + self._read_exhausted_raw()
        else:
            while len(self._read_buffer) < size:
                fill_size = self._fill_new_bytes()
                if fill_size == 0:
                    return self._read_buffer.read()

            return self._read_buffer.read(size)

    def _read_exhausted_raw(self) -> bytes:
        raise NotImplementedError

    def _fill_new_bytes(self) -> int:
        raise NotImplementedError

    def close(self) -> None:
        self._read_buffer.close()


class HttpBufferedFileLike(BaseBufferedFileLike):
    def __init__(
        self,
        url: str,
        headers: t.Optional[t.Dict[str, str]] = None,
        timeout: int = 90,
        buffer_size: int = 256 * 1024,
    ) -> None:
        super().__init__(buffer_size)
        if headers is None:
            self.headers = {"Accept-Encoding": "identity"}
        else:
            self.headers = headers

        self.buffer_size = buffer_size
        self.resp = requests.get(
            url,
            stream=True,
            headers=self.headers,
            timeout=timeout,
        )

        if not self.resp.ok:
            self.resp.raise_for_status()

        self._read_iter = self.resp.iter_content(buffer_size)

    def read(self, size: int) -> bytes:
        if size == 0:
            return b""
        elif size < 0:
            # urllib3 resp read does not support to read -1
            return self.resp.content  # type: ignore
        else:
            return self.resp.raw.read(size)  # type: ignore

    def close(self) -> None:
        self.resp.close()
        return super().close()


class BytesBuffer:
    def __init__(self, size: int = io.DEFAULT_BUFFER_SIZE) -> None:
        self._size = size
        self._pos = 0
        self._buffer = bytearray(0)

    def __str__(self) -> str:
        return f"buffer size: {self._size}, pos: {self._pos}, length: {len(self)}"

    __repr__ = __str__

    def close(self) -> None:
        self._buffer = bytearray(0)

    def __len__(self) -> int:
        return len(self._buffer) - self._pos

    def read(self, size: int = -1) -> bytes:
        if size < 0 or size > len(self):
            size = len(self)

        ret = self._buffer[self._pos : self._pos + size]
        self._pos += size
        return bytes(ret)

    def write(self, new_bytes: bytes) -> int:
        self._buffer = self._buffer[self._pos :] + new_bytes
        self._pos = 0
        return len(new_bytes)

    def write_from_iter(self, source: t.Any) -> int:
        # TODO: add iterable check for source
        new_bytes = bytearray()
        for c in source:
            new_bytes += c
            if len(new_bytes) >= self._size:
                break
        return self.write(new_bytes)


class S3BufferedFileLike(BaseBufferedFileLike):
    # TODO: add s3 typing
    def __init__(self, s3: t.Any, bucket: str, key: str, start: int, end: int) -> None:
        super().__init__(_CHUNK_SIZE)
        self.key = self._format_key(key, bucket)
        self.obj = s3.Object(bucket, self.key)

        self.end = end
        self._s3_eof = False
        self._current_s3_start = start

    def _format_key(self, key: str, bucket: str) -> str:
        r = urlparse(key)
        path = r.path.lstrip("/")
        if r.netloc:
            _flag = f"{bucket}/"
            if not path.startswith(_flag):
                raise FormatError(f"{key} does not contain bucket({bucket}) prefix")
            return path.split(_flag, 1)[-1]
        else:
            return path

    def _read_exhausted_raw(self) -> bytes:
        return self.obj.get()["Body"].read()  # type: ignore

    def _fill_new_bytes(self) -> int:
        new_bytes, _ = self._next_data()
        return self._read_buffer.write(new_bytes)

    def _next_data(self) -> t.Tuple[bytes, int]:
        end = _CHUNK_SIZE + self._current_s3_start - 1
        end = end if self.end == FilePosition.END else min(self.end, end)

        data, length = self._do_fetch_data(self._current_s3_start, end)
        self._current_s3_start += length

        return data, length

    def _do_fetch_data(self, _start: int, _end: int) -> t.Tuple[bytes, int]:
        # TODO: add more exception handle
        if self._s3_eof or (_end != FilePosition.END and _end < _start):
            return b"", 0

        resp = self.obj.get(Range=f"bytes={_start}-{_end}")
        body = resp["Body"]
        length = resp["ContentLength"]
        out = resp["Body"].read()
        body.close()

        self._s3_eof = _end == FilePosition.END or (_end - _start + 1) > length
        return out, length



import sys
import typing as t
import importlib
from pathlib import Path

import pkg_resources
from loguru import logger

from starwhale.utils import console
from starwhale.utils.venv import (
    guess_current_py_env,
    get_user_python_sys_paths,
    check_python_interpreter_consistency,
)


def import_object(
    workdir: t.Union[Path, str], handler_path: str, py_env: str = ""
) -> t.Any:
    workdir_path = str(Path(workdir).absolute())
    external_paths = [workdir_path]
    py_env = py_env or guess_current_py_env()
    _ok, _cur_py, _ex_py = check_python_interpreter_consistency(py_env)
    if not _ok:
        console.print(
            f":speaking_head: [red]swcli python prefix:{_cur_py}, runtime env python prefix:{_ex_py}[/], swcli will inject sys.path"
        )
        external_paths.extend(get_user_python_sys_paths(py_env))

    prev_paths = sys.path[:]
    sys_changed = False

    for _path in external_paths[::-1]:
        if _path not in sys.path:
            logger.debug(f"insert sys.path: '{_path}'")
            sys.path.insert(0, _path)
            pkg_resources.working_set.add_entry(_path)
            sys_changed = True

    try:
        module_name, handler_name = handler_path.split(":", 1)
        logger.debug(f"import module:{module_name}, handler:{handler_name}")
        _module = importlib.import_module(module_name, package=workdir_path)
        _obj = getattr(_module, handler_name, None)

        if not _obj:
            raise ModuleNotFoundError(f"{handler_path}")
    except Exception as e:
        logger.exception(e)
        if sys_changed:
            sys.path[:] = prev_paths
        raise

    return _obj


def load_module(module: str, path: Path) -> t.Any:
    workdir_path = str(path.absolute())

    external_paths = [workdir_path]
    for _path in external_paths[::-1]:
        if _path not in sys.path:
            logger.debug(f"insert sys.path: '{_path}'")
            sys.path.insert(0, _path)
            pkg_resources.working_set.add_entry(_path)

    return importlib.import_module(module, package=workdir_path)



import typing as t

import numpy
import torch

__all__ = ["convert_list_to_tensor", "convert_numpy_to_tensor"]


def convert_list_to_tensor(data: t.List) -> torch.Tensor:
    if not isinstance(data, list):
        raise TypeError(f"data is not list type: {data}")

    return torch.Tensor(data)


def convert_numpy_to_tensor(data: numpy.ndarray) -> torch.Tensor:
    if not isinstance(data, numpy.ndarray):
        raise TypeError(f"data is not numpy.ndarray type: {data}")

    return torch.as_tensor(data)



class SWEnv:
    instance_uri = "SW_INSTANCE_URI"
    project = "SW_PROJECT"
    task_id = "SW_TASK_ID"
    eval_version = "SW_EVALUATION_VERSION"
    model_version = "SW_MODEL_VERSION"
    runtime_version = "SW_RUNTIME_VERSION"
    status_dir = "SW_TASK_STATUS_DIR"
    log_dir = "SW_TASK_LOG_DIR"
    dataset_uri = "SW_DATASET_URI"
    dataset_row_start = "SW_DATASET_ROW_START"
    dataset_row_end = "SW_DATASET_ROW_END"
    instance_token = "SW_TOKEN"



from __future__ import annotations

import os
import copy
import typing as t
import tempfile
from http import HTTPStatus
from concurrent.futures import as_completed, ThreadPoolExecutor

import yaml
from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    TotalFileSizeColumn,
    TransferSpeedColumn,
)

from starwhale.utils import console, now_str, load_yaml
from starwhale.consts import (
    FileDesc,
    HTTPMethod,
    CREATED_AT_KEY,
    SHORT_VERSION_CNT,
    STANDALONE_INSTANCE,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_file
from starwhale.base.type import InstanceType
from starwhale.utils.error import NotFoundError, NoSupportError
from starwhale.base.bundle_copy import BundleCopy, _UploadPhase
from starwhale.core.dataset.store import DatasetStorage

from .tabular import TabularDataset, DatastoreRevision

# local storage scheme: no-scheme or file://
_LOCAL_STORAGE_SCHEMES = ("", "file")


class DatasetCopy(BundleCopy):
    def __init__(self, src_uri: str, dest_uri: str, typ: str, **kw: t.Any) -> None:
        super().__init__(src_uri, dest_uri, typ, **kw)
        self._max_workers = int(os.environ.get("SW_BUNDLE_COPY_THREAD_NUM", "5"))

    def _check_dataset_existed(self, uri: URI) -> bool:
        dataset_name = uri.object.name or self.bundle_name
        if uri.instance_type == InstanceType.CLOUD:
            ok, _ = self.do_http_request_simple_ret(
                path=f"/project/{uri.project}/dataset/{dataset_name}",
                method=HTTPMethod.HEAD,
                instance_uri=uri,
                ignore_status_codes=[HTTPStatus.NOT_FOUND],
            )
            return ok
        else:
            dataset_dir = (
                self._sw_config.rootdir / uri.project / "dataset" / dataset_name
            )
            return (dataset_dir / DEFAULT_MANIFEST_NAME).exists()

    def do(self) -> None:
        if not self._check_version_existed(self.src_uri):
            raise NotFoundError(f"src dataset not found: {self.src_uri}")

        # TODO: add `--overwrite` for dataset copy
        if self._check_dataset_existed(self.dest_uri):
            raise NoSupportError(
                f"dest dataset already exists, no support to overwrite or update copy: {self.dest_uri}"
            )

        console.print(f":construction: start to copy {self.src_uri} -> {self.dest_uri}")

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeElapsedColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            console=console,
            refresh_per_second=0.2,
        ) as progress:
            src = TabularDataset(
                name=self.src_uri.object.name,
                project=self.src_uri.project,
                instance_name=self.src_uri.instance,
            )

            dest = TabularDataset(
                name=self.dest_uri.object.name or self.src_uri.object.name,
                project=self.dest_uri.project,
                instance_name=self.dest_uri.instance,
            )
            try:
                self._do_dataset_copy(src=src, dest=dest, progress=progress)
            finally:
                src.close()
                dest.close()

        remote_url = self._get_remote_bundle_console_url()
        console.print(f":tea: console url of the remote bundle: {remote_url}")

    def _do_dataset_copy(
        self,
        src: TabularDataset,
        dest: TabularDataset,
        progress: Progress,
    ) -> None:
        console.print(":bird: preprocess artifacts link")
        _artifacts_uri_map: t.Dict[str, str] = {}

        for row in src.scan():
            for artifact in row.artifacts:
                link = artifact.link
                if (
                    not link
                    or link.scheme not in _LOCAL_STORAGE_SCHEMES
                    or link.uri in _artifacts_uri_map
                ):
                    continue

                _artifacts_uri_map[link.uri] = ""

        if dest.instance_name == STANDALONE_INSTANCE:
            console.print(f":cat: try to download {len(_artifacts_uri_map)} blobs")
            self._do_download_blobs(_artifacts_uri_map, progress)
        else:
            console.print(f":cat: try to upload {len(_artifacts_uri_map)} blobs")
            self._do_upload_blobs(_artifacts_uri_map, progress)

        console.print(":horse: dump dataset meta from src to dest")
        for row in src.scan():
            for artifact in row.artifacts:
                link = artifact.link
                if (
                    not link
                    or link.scheme not in _LOCAL_STORAGE_SCHEMES
                    or link.uri not in _artifacts_uri_map
                ):
                    continue
                link.uri = _artifacts_uri_map[link.uri]
            dest.put(row)

        console.print(":kangaroo: update dataset info")
        dest._info = copy.deepcopy(src.info)

        dataset_revision, info_revision = dest.flush()
        console.print(":tiger: make version for dest instance")
        if dest.instance_name == STANDALONE_INSTANCE:
            self._make_standalone_version(dataset_revision, info_revision)
        else:
            self._make_cloud_version(dataset_revision, info_revision)

    def _make_standalone_version(
        self, dataset_revision: str, info_revision: str
    ) -> None:
        r = self.do_http_request(
            path=f"/project/{self.src_uri.project}/dataset/{self.src_uri.object.name}",
            instance_uri=self.src_uri,
            params={"versionUrl": self.src_uri.object.version},
        ).json()

        manifest = yaml.safe_load(r["data"]["versionMeta"])
        manifest[CREATED_AT_KEY] = now_str()
        manifest.update(
            DatastoreRevision(data=dataset_revision, info=info_revision).asdict()
        )

        _dest_uri = URI.capsulate_uri(
            instance=STANDALONE_INSTANCE,
            project=self.dest_uri.project,
            obj_type=self.typ,
            obj_name=self.dest_uri.object.name or self.bundle_name,
            obj_ver=self.bundle_version,
        )
        snapshot_dir = self._get_versioned_resource_path(_dest_uri)
        ensure_file(
            snapshot_dir / DEFAULT_MANIFEST_NAME, yaml.safe_dump(manifest), parents=True
        )

        StandaloneTag(_dest_uri).add_fast_tag()

    def _make_cloud_version(self, dataset_revision: str, info_revision: str) -> None:
        dataset_name = self.dest_uri.object.name or self.src_uri.object.name
        params = {
            "swds": f"{dataset_name}:{self.bundle_name}",
            "project": self.dest_uri.project,
            "force": "1",  # use force=1 to make http retry happy, we check dataset existence in advance
        }
        url_path = self._get_remote_bundle_api_url()
        snapshot_dir = self._get_versioned_resource_path(self.src_uri)
        manifest = load_yaml(snapshot_dir / DEFAULT_MANIFEST_NAME)
        manifest[CREATED_AT_KEY] = now_str()
        manifest.update(
            DatastoreRevision(data=dataset_revision, info=info_revision).asdict()
        )
        _, tmp_path = tempfile.mkstemp()
        try:
            ensure_file(tmp_path, yaml.dump(manifest), parents=True)

            # TODO: use dataset create api
            r = self.do_multipart_upload_file(
                url_path=url_path,
                file_path=tmp_path,
                instance_uri=self.dest_uri,
                params={
                    "phase": _UploadPhase.MANIFEST,
                    "desc": FileDesc.MANIFEST.name,
                    **params,
                },
                use_raise=True,
            )
            self.do_http_request(
                path=url_path,
                method=HTTPMethod.POST,
                instance_uri=self.dest_uri,
                data={
                    "phase": _UploadPhase.END,
                    "uploadId": r.json()["data"]["uploadId"],
                    **params,
                },
                use_raise=True,
                disable_default_content_type=True,
            )
        finally:
            os.unlink(tmp_path)

    def _do_download_blobs(
        self,
        artifacts_uri_map: t.Dict[str, str],
        progress: Progress,
    ) -> None:
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            futures = [
                executor.submit(
                    self._do_download_blob_to_object_store,
                    progress,
                    k,
                )
                for k, v in artifacts_uri_map.items()
            ]

            results = [t.result() for t in as_completed(futures)]

            for src_uri, dest_uri in results:
                artifacts_uri_map[src_uri] = dest_uri

    def _do_download_blob_to_object_store(
        self,
        progress: Progress,
        remote_uri: str,
    ) -> t.Tuple[str, str]:
        hash_name = remote_uri.strip().strip("/").split("/")[-1]

        # TODO: get size by head api
        task_id = progress.add_task(
            f":arrow_down: {hash_name[:SHORT_VERSION_CNT]}",
            visible=True,
        )

        local_blob_path = DatasetStorage._get_object_store_path(hash_name)
        if not local_blob_path.exists():
            self.do_download_file(
                url_path=f"/project/{self.src_uri.project}/dataset/{self.src_uri.object.name}/blob",
                dest_path=local_blob_path,
                instance_uri=self.src_uri,
                progress=progress,
                task_id=task_id,
            )
        else:
            progress.update(task_id, completed=local_blob_path.stat().st_size)

        return remote_uri, hash_name

    def _do_upload_blob_from_object_store(
        self, progress: Progress, local_hashed_uri: str
    ) -> t.Tuple[str, str]:
        local_blob_path = DatasetStorage._get_object_store_path(local_hashed_uri)
        url_path = f"/project/{self.dest_uri.project}/dataset/{self.bundle_name}/hashedBlob/{local_hashed_uri}"
        blob_size = local_blob_path.stat().st_size

        task_id = progress.add_task(
            f":arrow_up: {local_hashed_uri[:SHORT_VERSION_CNT]}",
            total=blob_size,
            visible=True,
        )

        r = self.do_http_request(
            path=url_path,
            instance_uri=self.dest_uri,
            method=HTTPMethod.HEAD,
            ignore_status_codes=[HTTPStatus.NOT_FOUND],
        )
        if r.status_code == HTTPStatus.OK:
            progress.update(task_id, completed=blob_size)
            return local_hashed_uri, r.headers["X-SW-LOCAL-STORAGE-URI"]

        remote_uri = self.do_multipart_upload_file(
            url_path=url_path,
            file_path=local_blob_path,
            instance_uri=self.dest_uri,
            use_raise=True,
            progress=progress,
            task_id=task_id,
        ).json()["data"]
        return local_hashed_uri, remote_uri

    def _do_upload_blobs(
        self,
        artifacts_uri_map: t.Dict[str, str],
        progress: Progress,
    ) -> None:

        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            futures = [
                executor.submit(
                    self._do_upload_blob_from_object_store,
                    progress,
                    k,
                )
                for k, v in artifacts_uri_map.items()
            ]

            results = [t.result() for t in as_completed(futures)]

            for src_uri, dest_uri in results:
                artifacts_uri_map[src_uri] = dest_uri



from __future__ import annotations

import typing as t
import logging

import loguru


class StreamWrapper:
    is_wrapper = True

    def __init__(
        self,
        stream: t.TextIO,
        logger: loguru.Logger,
        level: int = logging.INFO,
    ) -> None:
        self.stream = stream
        self.logger = logger
        self.level = level

    def write(self, sb: bytes) -> None:
        s = str(sb)
        self.stream.write(s)
        # TODO: splitlines?
        self.logger.opt(depth=1).log(self.level, s.rstrip())

    def __getattr__(self, name: str) -> t.Any:
        return getattr(self.stream, name)

    def flush(self) -> None:
        ...



import os
from typing import Iterator
from pathlib import Path

from starwhale.utils import load_yaml
from starwhale.consts import FileDesc, FileNode, SWMP_SRC_FNAME, DEFAULT_MANIFEST_NAME
from starwhale.utils.fs import extract_tar
from starwhale.base.bundle_copy import BundleCopy


class ModelCopy(BundleCopy):
    def upload_files(self, workdir: Path) -> Iterator[FileNode]:
        _manifest = load_yaml(workdir / DEFAULT_MANIFEST_NAME)
        for _m in _manifest.get("resources", []):
            if _m["desc"] != FileDesc.MODEL.name:
                continue
            _path = workdir / _m["path"]
            yield FileNode(
                path=_path,
                name=_m["name"],
                size=_path.stat().st_size,
                file_desc=FileDesc.MODEL,
                signature=_m["signature"],
            )

        _meta_names = [SWMP_SRC_FNAME]

        for _n in _meta_names:
            _path = workdir / _n
            yield FileNode(
                path=_path,
                name=os.path.basename(_path),
                size=_path.stat().st_size,
                file_desc=FileDesc.SRC_TAR,
                signature="",
            )

    def download_files(self, workdir: Path) -> Iterator[FileNode]:
        _manifest = load_yaml(workdir / DEFAULT_MANIFEST_NAME)
        for _f in (SWMP_SRC_FNAME,):
            yield FileNode(
                path=workdir / _f,
                signature="",
                size=0,
                name=_f,
                file_desc=FileDesc.SRC_TAR,
            )
            extract_tar(
                tar_path=workdir / SWMP_SRC_FNAME,
                dest_dir=workdir / "src",
                force=False,
            )
        # this must after src download
        for _m in _manifest.get("resources", []):
            if _m["desc"] != FileDesc.MODEL.name:
                continue
            _sign = _m["signature"]
            _dest = workdir / _m["path"]

            yield FileNode(
                path=_dest,
                signature=_sign,
                size=0,
                name=_m["name"],
                file_desc=FileDesc.MODEL,
            )
            # TODO use unified storage
            # Path(workdir / _m["path"]).symlink_to(
            #     _dest # the unify dir
            # )



from typing import Tuple

import click

from starwhale.utils.cli import AliasedGroup

from .host import CommandRetriever
from .remote import CommandRunner


@click.group("assistance", cls=AliasedGroup, help="Remote assistance")
def assistance_cmd() -> None:
    pass


@assistance_cmd.command(
    "host", help="Run a remote assistance host to accept remote commands"
)
@click.option("--broker", help="Broker URL", required=True)
def host(broker: str) -> None:
    retriever = CommandRetriever("host", broker)
    retriever.start()
    try:
        retriever.join()
    except KeyboardInterrupt:
        print("stopping...")
        retriever.stop()
        retriever.join()


@assistance_cmd.command(
    "remote",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
    help="Send commands to remote assistance host",
)
@click.option("--broker", help="Broker URL", required=True)
@click.argument("args", nargs=-1, required=True)
def remote(broker: str, args: Tuple[str]) -> None:
    runner = CommandRunner("remote", broker, list(args))
    runner.start()
    try:
        runner.join()
    except KeyboardInterrupt:
        print("stopping...")
        runner.stop()
        runner.join()



import sys
import typing as t

from rich import box
from loguru import logger
from rich.panel import Panel
from rich.table import Table
from rich.pretty import Pretty
from rich.columns import Columns

from starwhale.utils import Order, console, sort_obj_list
from starwhale.consts import (
    CREATED_AT_KEY,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
    DEFAULT_REPORT_COLS,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType, JobOperationType
from starwhale.base.view import BaseTermView
from starwhale.core.eval.model import EvaluationJob
from starwhale.api._impl.metric import MetricKind
from starwhale.base.uricomponents.instance import Instance


class JobTermView(BaseTermView):
    def __init__(self, job_uri: str) -> None:
        super().__init__()
        self.raw_uri = job_uri
        self.uri = URI(job_uri, expected_type=URIType.EVALUATION)
        logger.debug(f"eval job:{self.raw_uri}")
        self.job = EvaluationJob.get_job(self.uri)
        self._action_run_map = {
            JobOperationType.CANCEL: self.job.cancel,
            JobOperationType.RESUME: self.job.resume,
            JobOperationType.REMOVE: self.job.remove,
            JobOperationType.RECOVER: self.job.recover,
            JobOperationType.PAUSE: self.job.pause,
        }

    def recover(self, force: bool = False) -> None:
        self._do_action(JobOperationType.RECOVER, force)

    def remove(self, force: bool = False) -> None:
        self._do_action(JobOperationType.REMOVE, force)

    def cancel(self, force: bool = False) -> None:
        self._do_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> None:
        self._do_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> None:
        self._do_action(JobOperationType.PAUSE, force)

    @BaseTermView._simple_action_print
    def _do_action(self, action: str, force: bool = False) -> t.Tuple[bool, str]:
        return self._action_run_map[action](force)

    @BaseTermView._only_standalone
    @BaseTermView._header
    def compare(self, job_uris: t.List[str]) -> None:
        if self.uri.instance_type != InstanceType.STANDALONE:
            console.print(
                ":dragon_face: Today job compare only works for standalone instance"
            )
            sys.exit(1)

        jobs = []
        for _u in job_uris:
            _uri = URI(_u, expected_type=URIType.EVALUATION)
            jobs.append(EvaluationJob.get_job(_uri))

        rt = self.job.compare(jobs)
        table = Table(
            box=box.ASCII,
            title=f":leafy_green: {rt['kind'].upper()} Job Compare Summary",
            caption=f":cactus: base job:{rt['base']['uri']}",
            expand=True,
        )
        table.add_column("", justify="left", no_wrap=True, style="cyan")
        for _v in rt["versions"]:
            table.add_column(_v[:SHORT_VERSION_CNT])

        for _k, _vs in rt["summary"].items():
            _ts = [_k]

            _is_empty = all([_v["value"] is None for _v in _vs])
            if _is_empty:
                continue

            for _v in _vs:
                if _v["base"]:
                    _ts.append(f":triangular_flag: {_v['value']}")
                else:
                    if _v["delta"] is None:
                        _ts.append(_v["value"])
                    else:
                        if _v["delta"] > 0:
                            _text = (
                                f":chart_with_upwards_trend: [green]{_v['value']}[/]"
                            )
                        elif _v["delta"] < 0:
                            _text = (
                                f":chart_with_downwards_trend: [red]{_v['value']}[/]"
                            )
                        else:
                            _text = f":white_circle: {_v['value']}"
                        _ts.append(_text)

            table.add_row(*([str(_t) for _t in _ts]))

        console.print(table)

    @BaseTermView._header
    def info(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        max_report_cols: int = DEFAULT_REPORT_COLS,
        web: bool = False,
    ) -> None:
        _rt = self.job.info(page, size)
        if not _rt:
            console.print(":tea: not found info")
            return

        manifest = _rt.get("manifest", {})
        if web:
            from starwhale.web.server import Server

            ver = manifest.get("id")

            # if id is numeric, it's a remote eval
            if ver and ver.isnumeric():
                svr = Server.proxy(Instance())
            else:
                # local eval
                if not manifest.get("version"):
                    console.print(":tea: eval id not found")
                    sys.exit(1)
                ver = manifest.get("version")
                svr = Server.default()

            # TODO support changing host and port
            host = "127.0.0.1"
            port = 8000
            url = f"http://{host}:{port}/projects/{self.uri.project}/evaluations/{ver}/results?token=local"
            console.print(f":tea: open {url} in browser")
            import uvicorn

            uvicorn.run(svr, host=host, port=port, log_level="error")
            return

        if manifest:
            console.rule(
                f"[green bold]Inspect {DEFAULT_MANIFEST_NAME} for eval:{self.uri}"
            )
            console.print(Pretty(_rt["manifest"], expand_all=True))

        if "location" in _rt:
            console.rule("Evaluation process dirs")
            console.print(f":cactus: ppl: {_rt['location']['ppl']}")
            console.print(f":camel: cmp: {_rt['location']['cmp']}")

        if "tasks" in _rt:
            self._print_tasks(_rt["tasks"][0])

        if "report" in _rt:
            _report = _rt["report"]
            _kind = _rt["report"].get("kind", "")

            if "summary" in _report:
                self._render_summary_report(_report["summary"], _kind)

            if _kind == MetricKind.MultiClassification.value:
                self._render_multi_classification_job_report(
                    _rt["report"], max_report_cols
                )

    def _render_summary_report(self, summary: t.Dict[str, t.Any], kind: str) -> None:
        console.rule(f"[bold green]{kind.upper()} Summary")
        contents = [
            Panel(f"[b]{k}[/b]\n[yellow]{v}", expand=True) for k, v in summary.items()
        ]
        console.print(Columns(contents))

    def _print_tasks(self, tasks: t.List[t.Dict[str, t.Any]]) -> None:
        table = Table(box=box.SIMPLE)
        table.add_column("ID", justify="left", style="cyan", no_wrap=True)
        table.add_column("UUID")
        table.add_column("Status", style="magenta")
        table.add_column("Duration")
        table.add_column("Created")
        table.add_column("Finished")

        for _t in tasks:
            status, style, icon = self.pretty_status(_t["taskStatus"])
            table.add_row(
                _t["id"],
                _t["uuid"],
                f"[{style}]{icon}{status}[/]",
                "",
                _t[CREATED_AT_KEY],
                "",
            )

        console.rule(
            f"[bold green]Project({self.uri.project} Job({self.job.name}) Tasks List"
        )
        console.print(table)

    def _render_multi_classification_job_report(
        self, report: t.Dict[str, t.Any], max_report_cols: int
    ) -> None:
        if not report:
            console.print(":turtle: no report")
            return

        labels: t.Dict[str, t.Any] = report.get("labels", {})
        sort_label_names = sorted(list(labels.keys()))

        def _print_labels() -> None:
            if len(labels) == 0:
                return

            table = Table(box=box.SIMPLE)
            table.add_column("Label", style="cyan")
            keys = labels[sort_label_names[0]]
            for idx, _k in enumerate(keys):
                if _k == "id":
                    continue
                # add 1: because "id" is first col and should be removed
                if idx < max_report_cols + 1:
                    table.add_column(_k.capitalize())
                else:
                    table.add_column("...")
                    break

            for _k, _v in labels.items():
                table.add_row(
                    _k,
                    *(
                        f"{float(_v[_k2]):.4f}"
                        for idx, _k2 in enumerate(keys)
                        if _k2 != "id" and idx < max_report_cols + 1
                    ),
                )

            console.rule(f"[bold green]{report['kind'].upper()} Label Metrics Report")
            console.print(table)

        def _print_confusion_matrix() -> None:
            cm = report.get("confusion_matrix", {})
            if not cm:
                return

            btable = Table(box=box.ROUNDED)
            btable.add_column("", style="cyan")
            for idx, n in enumerate(sort_label_names):
                if idx < max_report_cols:
                    btable.add_column(n)
                else:
                    btable.add_column("...")
                    break
            for idx, bl in enumerate(cm.get("binarylabel", [])):
                btable.add_row(
                    sort_label_names[idx],
                    *[
                        f"{float(bl[i]):.4f}"
                        for idx, i in enumerate(bl)
                        if i != "id" and idx < max_report_cols + 1
                    ],
                )
            console.rule(f"[bold green]{report['kind'].upper()} Confusion Matrix")
            console.print(btable)

        _print_labels()
        _print_confusion_matrix()

    @classmethod
    def run(
        cls,
        project_uri: str,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        version: str = "",
        name: str = "",
        desc: str = "",
        step_spec: str = "",
        resource_pool: str = "",
        gencmd: bool = False,
        use_docker: bool = False,
        step: str = "",
        task_index: int = 0,
        task_num: int = 0,
    ) -> str:
        _project_uri = URI(project_uri, expected_type=URIType.PROJECT)
        ok, version = EvaluationJob.run(
            _project_uri,
            model_uri,
            dataset_uris,
            runtime_uri,
            version=version,
            name=name,
            desc=desc,
            step_spec=step_spec,
            resource_pool=resource_pool,
            gencmd=gencmd,
            use_docker=use_docker,
            step=step,
            task_index=task_index,
            task_num=task_num,
        )

        # TODO: show report in standalone mode directly

        if ok:
            console.print(
                f":clap: success to create job(project id: [red]{_project_uri.full_uri}[/])"
            )
            if _project_uri.instance_type == InstanceType.CLOUD:
                _job_uri = f"{_project_uri.full_uri}/evaluation/{version}"
            else:
                _job_uri = f"{version[:SHORT_VERSION_CNT]}"

            console.print(
                f":bird: run cmd to fetch eval info: [bold green]swcli eval info {_job_uri}[/]"
            )
            return version
        else:
            console.print(
                f":collision: failed to create eval job, notice: [red]{version}[/]"
            )
            return ""

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        cls.must_have_project(_uri)
        jobs, pager = EvaluationJob.list(_uri, page, size)
        jobs = sort_obj_list(jobs, [Order("manifest.created_at", True)])
        return (jobs, pager)


class JobTermViewRich(JobTermView):
    @classmethod
    @BaseTermView._pager
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        _jobs, _pager = super().list(project_uri, fullname, show_removed, page, size)
        table = Table(title="Job List", box=box.SIMPLE, expand=True)
        table.add_column("Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Model", no_wrap=True)
        table.add_column("Datasets")
        table.add_column("Runtime")
        table.add_column("Status", style="red")
        table.add_column("Resource", style="blue")
        table.add_column("Created At", style="magenta")
        table.add_column("Finished At", style="magenta")

        def _s(x: str) -> str:
            _end = -1 if fullname else SHORT_VERSION_CNT
            if ":" in x:
                _n, _v = x.split(":")
                return f"{_n}:{_v[:_end]}"
            else:
                return x[:_end]

        for _job in _jobs:
            if show_removed ^ _job.get("is_removed", False):
                continue

            _m = _job["manifest"]
            _status, _style, _icon = cls.pretty_status(
                _m.get("jobStatus") or _m.get("status")
            )
            _datasets = "--"
            if _m.get("datasets"):
                _datasets = "\n".join([_s(d) for d in _m["datasets"]])

            _model = "--"
            if "model" in _m:
                _model = _s(_m["model"])
            elif "modelName" in _m:
                _model = f"{_m['modelName']}:{_s(_m['modelVersion'])}"

            _name = "--"
            if "id" in _m:
                _name = _m["id"]
            else:
                _name = _m["version"] if show_removed else _s(_m["version"])

            _runtime = "--"
            if "runtime" in _m:
                if isinstance(_m["runtime"], str):
                    _runtime = _m["runtime"]
                else:
                    _r = _m["runtime"]
                    _runtime = f"[{_r['version']['id']}] {_r['name']}:{_r['version']['name'][:SHORT_VERSION_CNT]}"

            table.add_row(
                _name,
                _model,
                _datasets,
                _runtime,
                f"[{_style}]{_icon}{_status}[/]",
                f"{_m['device']}:{_m['deviceAmount']}" if "device" in _m else "--",
                _m[CREATED_AT_KEY],
                _m["finished_at"],
            )
            # TODO: add duration
        cls.print_header()
        console.print(table)
        return _jobs, _pager


class JobTermViewJson(JobTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        _data, _ = super().list(project_uri, fullname, show_removed, page, size)
        cls.pretty_json(_data)

    def info(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        max_report_cols: int = DEFAULT_REPORT_COLS,
        web: bool = False,
    ) -> None:
        _rt = self.job.info(page, size)
        if not _rt:
            console.print(":tea: not found info")
            return
        self.pretty_json(_rt)


def get_term_view(ctx_obj: t.Dict) -> t.Type[JobTermView]:
    return JobTermViewJson if ctx_obj.get("output") == "json" else JobTermViewRich



import threading
from typing import Any, Dict, List, Optional
from datetime import datetime

from fastapi import Body, FastAPI, Request, Response, HTTPException
from pydantic import BaseSettings

from .common import (
    random_id,
    ChunkBuffer,
    ClosedError,
    PeriodicRunner,
    InvalidOffsetError,
)


class Settings(BaseSettings):
    broker_url: str = ""
    chunk_size: int = 65536
    max_chunk_count: int = 64
    gc_timeout_seconds = 600.0
    gc_interval_seconds = 60.0
    keep_alive_seconds = 15.0
    keep_alive_check_interval_seconds = 5.0


settings = Settings()
app = FastAPI()
global_lock = threading.Lock()


class Command:
    def __init__(self, args: List[str]):
        self.args = args
        self.exit_code: Optional[int] = None
        self.files: Dict[str, ChunkBuffer] = {
            "stdin": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
            "stdout": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
            "stderr": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
        }
        self.last_access_time = datetime.now().timestamp()
        self.host_access_time = self.last_access_time
        self.remote_access_time = self.last_access_time

    def get_file(self, filename: str) -> ChunkBuffer:
        if filename not in self.files:
            raise HTTPException(status_code=404, detail=f"invalid filename {filename}")
        return self.files[filename]


class Session:
    def __init__(self) -> None:
        self.commands: Dict[str, Command] = {}
        self.last_access_time = datetime.now().timestamp()

    def new_command(self, command_id: str, args: List[str]) -> None:
        if command_id not in self.commands:
            self.commands[command_id] = Command(args)

    def get_command(self, command_id: str) -> Command:
        if command_id not in self.commands:
            raise HTTPException(
                status_code=404, detail=f"invalid command id {command_id}"
            )
        command = self.commands[command_id]
        command.last_access_time = datetime.now().timestamp()
        return command

    def list_commands(self) -> List[Dict[str, Any]]:
        ret = []
        for command_id, command in self.commands.items():
            ret.append(
                {
                    "command_id": command_id,
                    "args": command.args,
                    "exitCode": command.exit_code,
                }
            )
        return ret


class SessionManager:
    def __init__(self) -> None:
        self.sessions: Dict[str, Session] = {}
        self.garbage_collector: Optional[PeriodicRunner] = None
        self.keep_alive_monitor: Optional[PeriodicRunner] = None

    def keep_alive_check(self) -> None:
        with global_lock:
            now = datetime.now().timestamp()
            for session in self.sessions.values():
                for command in session.commands.values():
                    if now - command.host_access_time > settings.keep_alive_seconds:
                        command.get_file("stdin").close("host timeout")
                        command.get_file("stdout").close("host timeout")
                        command.get_file("stderr").close("host timeout")
                    if now - command.remote_access_time > settings.keep_alive_seconds:
                        command.get_file("stdin").close("remote timeout")
                        command.get_file("stdout").close("remote timeout")
                        command.get_file("stderr").close("remote timeout")

    def garbage_collect(self) -> None:
        with global_lock:
            now = datetime.now().timestamp()
            new_sessions = {}
            for session_id, session in self.sessions.items():
                if now - session.last_access_time > settings.gc_timeout_seconds:
                    print(f"garbage collection: remove session {session_id}")
                else:
                    new_sessions[session_id] = session
                    new_commands = {}
                    for command_id, command in session.commands.items():
                        if now - command.last_access_time > settings.gc_timeout_seconds:
                            print(f"garbage collection: remove command {command_id}")
                        else:
                            new_commands[command_id] = command
                    session.commands = new_commands
            self.sessions = new_sessions

    def start_garbage_collector(self) -> None:
        with global_lock:
            if self.garbage_collector is None:
                self.garbage_collector = PeriodicRunner(
                    "broker-garbage-collector",
                    self.garbage_collect,
                    settings.gc_interval_seconds,
                )
                self.garbage_collector.start()
                print("garbage collector started")

    def stop_garbage_collector(self) -> None:
        with global_lock:
            if self.garbage_collector is not None:
                self.garbage_collector.stop()
                self.garbage_collector = None

    def start_keep_alive_monitor(self) -> None:
        with global_lock:
            if self.keep_alive_monitor is None:
                self.keep_alive_monitor = PeriodicRunner(
                    "broker-keep-alive-monitor",
                    self.keep_alive_check,
                    settings.keep_alive_check_interval_seconds,
                )
                self.keep_alive_monitor.start()
                print("keep alive monitor started")

    def stop_keep_alive_monitor(self) -> None:
        with global_lock:
            if self.keep_alive_monitor is not None:
                self.keep_alive_monitor.stop()
                self.keep_alive_monitor = None

    def new_session(self) -> str:
        session_id = random_id()
        self.sessions[session_id] = Session()
        return session_id

    def get_session(self, session_id: str) -> Session:
        if session_id not in self.sessions:
            raise HTTPException(
                status_code=404, detail=f"invalid session id {session_id}"
            )
        session = self.sessions[session_id]
        session.last_access_time = datetime.now().timestamp()
        return session


session_manager = SessionManager()


@app.on_event("startup")
def startup_event() -> None:
    session_manager.start_garbage_collector()
    session_manager.start_keep_alive_monitor()


@app.post("/session")
def new_session() -> Dict[str, Any]:
    with global_lock:
        session_id = session_manager.new_session()
        return {
            "hostCmd": f"swcli assistance host --broker={settings.broker_url}/session/{session_id}",
            "remoteCmd": f"swcli assistance remote --broker={settings.broker_url}/session/{session_id} [args]",
        }


@app.post("/session/{session_id}/command")
def run_command(
    session_id: str, command_id: str = Body(), args: List[str] = Body()
) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        session.new_command(command_id, args)


@app.post("/session/{session_id}/command/{command_id}")
def update_command_exit_code(session_id: str, command_id: str, exit_code: int) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        command.exit_code = exit_code


@app.post("/session/{session_id}/command/{command_id}/file/{filename}")
async def write_command_file(
    session_id: str,
    command_id: str,
    filename: str,
    offset: int,
    request: Request,
) -> Dict[str, Any]:
    data = await request.body()
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
        if filename == "stdin":
            command.remote_access_time = datetime.now().timestamp()
        else:
            command.host_access_time = datetime.now().timestamp()
    try:
        count = f.write(offset, data, 1)
        return {"count": count}
    except (ClosedError, InvalidOffsetError) as e:
        raise HTTPException(400, str(e))


@app.post("/session/{session_id}/command/{command_id}/file/{filename}/closeRead")
def close_read_command_file(session_id: str, command_id: str, filename: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
    f.close("read close")


@app.post("/session/{session_id}/command/{command_id}/file/{filename}/closeWrite")
def close_write_command_file(session_id: str, command_id: str, filename: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
    f.close("")


@app.get(
    "/session/{session_id}/command/{command_id}/file/{filename}",
    response_class=Response,
    responses={200: {"content": {"application/octet-stream": {}}}},
)
def read_command_file(
    session_id: str, command_id: str, filename: str, offset: int
) -> Response:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
        if filename == "stdin":
            command.host_access_time = datetime.now().timestamp()
        else:
            command.remote_access_time = datetime.now().timestamp()
    try:
        data = f.read(offset, 1)
    except (ClosedError, InvalidOffsetError) as e:
        raise HTTPException(400, str(e))
    if data is None:
        return Response(
            headers={"x-wait-for-data": ""},
            content=b"",
            media_type="application/octet-stream",
        )
    return Response(content=data, media_type="application/octet-stream")


@app.post("/session/{session_id}/command/{command_id}/keepAlive")
def keep_alive(session_id: str, command_id: str, end: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        now = datetime.now().timestamp()
        if end == "host":
            command.host_access_time = now
        else:
            command.remote_access_time = now


@app.get("/session/{session_id}/command")
def list_commands(session_id: str) -> Dict[str, Any]:
    with global_lock:
        session = session_manager.get_session(session_id)
        return {"commands": session.list_commands()}



import typing

SuccessResp = typing.Dict[str, typing.Any]


def success(data: typing.Any) -> SuccessResp:
    return {
        "code": "success",
        "message": "Success",
        "data": data,
    }



from __future__ import annotations

import typing as t
from enum import Enum, unique
from functools import wraps

from sklearn.metrics import (  # type: ignore
    auc,
    roc_curve,
    hamming_loss,
    confusion_matrix,
    cohen_kappa_score,
    classification_report,
    multilabel_confusion_matrix,
)

from starwhale.api import evaluation
from starwhale.utils.dict import flatten as flatten_dict


@unique
class MetricKind(Enum):
    MultiClassification = "multi_classification"


def multi_classification(
    confusion_matrix_normalize: str = "all",
    show_hamming_loss: bool = True,
    show_cohen_kappa_score: bool = True,
    show_roc_auc: bool = True,
    all_labels: t.Optional[t.List[t.Any]] = None,
) -> t.Any:
    def _decorator(func: t.Any) -> t.Any:
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> t.Dict[str, t.Any]:
            y_pr: t.Any = None

            _rt = func(*args, **kwargs)
            if show_roc_auc:
                y_true, y_pred, y_pr = _rt
            else:
                y_true, y_pred = _rt

            _r: t.Dict[str, t.Any] = {"kind": MetricKind.MultiClassification.value}
            cr = classification_report(
                y_true, y_pred, output_dict=True, labels=all_labels
            )
            _summary_m = ["accuracy", "micro avg", "weighted avg", "macro avg"]
            _r["summary"] = {}
            for k in _summary_m:
                v = cr.get(k)
                if not v:
                    continue
                _r["summary"][k] = v

            if show_hamming_loss:
                _r["summary"]["hamming_loss"] = hamming_loss(y_true, y_pred)
            if show_cohen_kappa_score:
                _r["summary"]["cohen_kappa_score"] = cohen_kappa_score(y_true, y_pred)

            _record_summary = flatten_dict(_r["summary"], extract_sequence=True)
            _record_summary["kind"] = _r["kind"]

            evaluation.log_summary(_record_summary)

            _r["labels"] = {}
            mcm = multilabel_confusion_matrix(
                y_true, y_pred, labels=all_labels
            ).tolist()

            labels = all_labels or sorted([k for k in cr.keys() if k not in _summary_m])
            for _label, matrix in zip(labels, mcm):
                _label = str(_label)
                _report = cr.get(_label)
                if not _report:
                    continue

                _report.update(
                    {
                        "TN-True Negative": matrix[0][0],
                        "FP-False Positive": matrix[0][1],
                        "FN-False Negative": matrix[1][0],
                        "TP-True Positive": matrix[1][1],
                    }
                )

                _r["labels"][_label] = _report
                evaluation.log("labels", id=_label, metrics=_report)

            # TODO: tune performance, use intermediated result
            cm = confusion_matrix(
                y_true, y_pred, labels=all_labels, normalize=confusion_matrix_normalize
            )
            _cm_list = cm.tolist()
            _r["confusion_matrix"] = {"binarylabel": _cm_list}

            for _idx, _pa in enumerate(_cm_list):
                evaluation.log(
                    "confusion_matrix/binarylabel",
                    id=_idx,
                    metrics={str(_id): _v for _id, _v in enumerate(_pa)},
                )

            if show_roc_auc and all_labels is not None and y_true and y_pr:
                _r["roc_auc"] = {}
                for _idx, _label in enumerate(all_labels):
                    _ra_value = _calculate_roc_auc(y_true, y_pr, _label, _idx)
                    _r["roc_auc"][str(_label)] = _ra_value

                    for _id, (_fpr, _tpr, _threshold) in enumerate(
                        zip(_ra_value["fpr"], _ra_value["tpr"], _ra_value["thresholds"])
                    ):
                        evaluation.log(
                            f"roc_auc/{_label}",
                            id=_id,
                            metrics=dict(
                                tpr=_tpr,
                                fpr=_fpr,
                                threshold=_threshold,
                            ),
                        )

                        evaluation.log(
                            "labels", id=str(_label), metrics=dict(auc=_ra_value["auc"])
                        )
            return _r

        return _wrapper

    return _decorator


def _calculate_roc_auc(
    y_pred: t.List[t.Any],
    y_pr: t.List[t.Any],
    label: t.Any,
    idx: int,
) -> t.Dict[str, t.Any]:
    # TODO: add sample rate
    y_bin_true = [int(i == label) for i in y_pred]
    y_label_pr = [m[idx] for m in y_pr]
    fpr, tpr, thresholds = roc_curve(y_bin_true, y_label_pr)
    return dict(
        fpr=fpr.tolist(),
        tpr=tpr.tolist(),
        auc=auc(fpr, tpr),
        thresholds=thresholds.tolist(),
    )
